#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from keras import layers, backend
from keras.utils import conv_utils

from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer
from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH, QFloat


__all__ = ["PaddedConv2D", "QuantizedConv2D", "QuantizedConv2DTranspose"]


def apply_padding(inputs, strides, kernel_size, padding_value):
    """Apply "SAME" padding to the inputs

    Args:
        inputs (tf.Tensor): the inputs tensor.
        strides (tuple): the strides tuple.
        kernel_size (int): the kernel size.
        padding_value (tf.Tensor): the padding value to apply.

    Returns:
        tf.Tensor: padded inputs.
    """
    _, h, w, _ = inputs.shape
    filter_width = kernel_size[0]
    filter_height = kernel_size[1]
    if h % strides[0] == 0:
        pad_along_height = max(filter_height - strides[0], 0)
    else:
        pad_along_height = max(filter_height - (h % strides[0]), 0)
    if w % strides[1] == 0:
        pad_along_width = max(filter_width - strides[1], 0)
    else:
        pad_along_width = max(filter_width - (w % strides[1]), 0)
    pad_top = pad_along_height // 2
    pad_bottom = pad_along_height - pad_top
    pad_left = pad_along_width // 2
    pad_right = pad_along_width - pad_left
    padding = [[0, 0], [pad_top, pad_bottom], [pad_left, pad_right], [0, 0]]

    return tf.pad(inputs, padding, "CONSTANT", padding_value)


@tf.keras.utils.register_keras_serializable()
class PaddedConv2D(layers.Conv2D):
    """A convolutional layer that can use a custom padding value.

    Note that when a padding value is provided, padding 'SAME' will be applied with the provided
    value (overriding 'padding' parameter).

    Args:
        padding_value (float, optional): the value used when padding for the 'same' convolution
            type. If None, zero-padding is used. Defaults to None.
    """

    def __init__(self, *args, padding_value=None, **kwargs):
        super().__init__(*args, **kwargs)
        # Raise an error if the padding value is not a scalar
        if padding_value is not None:
            if tf.size(padding_value) != 1:
                raise ValueError(f"The padding value must be a scalar. Receives {padding_value}.")
            # When a custom padding_value is given, self.padding is overwritten and 'SAME' padding
            # will be done explicitly in the call.
            self._padding_value = tf.constant(padding_value)
            self.padding = 'valid'
        else:
            self._padding_value = None

    def call(self, inputs, training=None):
        # We need a custom padding for specifics padding values
        if self._padding_value is not None:
            # Note that we use the custom padding even when padding value is 0.
            inputs = apply_padding(inputs,
                                   list(self.strides),
                                   (self.kernel.shape[1], self.kernel.shape[0]),
                                   self._padding_value)

        outputs = self.convolution_op(inputs, self.kernel)

        if self.use_bias:
            outputs = tf.add(outputs, self.bias)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["padding_value"] = self._padding_value
        return config


@tf.keras.utils.register_keras_serializable()
class QuantizedConv2D(Calibrable, layers.Conv2D):
    """A convolutional layer that operates on quantized inputs and weights.

    Note that when a padding value is provided, padding 'SAME' will be applied with the provided
    value (overriding 'padding' parameter).

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
        padding_value (float, optional): the value used when padding for the 'same' convolution
            type. If None, zero-padding is used. Defaults to None.
    """

    # The padding_value frac_bits for the 'same' convolution type when we use
    # a tf.Tensor for this parameter.
    _PADDING_FRAC_BITS = 1

    def __init__(self, *args, padding_value=None, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.weight_quantizer = deserialize_quantizer(
            self.quant_config, "weight_quantizer", WeightQuantizer, True)
        if self.use_bias:
            self.bias_quantizer = deserialize_quantizer(
                self.quant_config, "bias_quantizer", WeightQuantizer, True)
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."
        if padding_value is not None:
            # Raise an error if the padding value is not a scalar
            if tf.size(padding_value) != 1:
                raise ValueError(f"The padding value must be a scalar. Receives {padding_value}.")
            padding_value = FixedPoint.quantize(padding_value,
                                                self.buffer_bitwidth,
                                                self._PADDING_FRAC_BITS)
            padding_value = padding_value.promote(self.buffer_bitwidth)
            # _padding_value is a private member because value controls are operated
            # in the constructor.
            # When a custom padding_value is given, self.padding is overwritten and 'SAME' padding
            # will be done explicitely in the call.
            self._padding_value = padding_value
            self.padding = 'valid'
        else:
            self._padding_value = None

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        if self._padding_value is not None:
            self.pad_shift = CalibrableVariable()
        if self.use_bias:
            self.conv_shift = CalibrableVariable()
            self.bias_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint or tf.Tensor
        if not isinstance(inputs, (FixedPoint, tf.Tensor)):
            raise TypeError(f"QuantizedConv2D only accepts FixedPoint or tf.Tensor\
                              inputs. Receives {type(inputs)} inputs.")

        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel, training)

        if isinstance(inputs, tf.Tensor):
            # Assume the inputs are integer stored as float, which is the only tf.Tensor
            # inputs that are allowed
            inputs = FixedPoint(inputs, 0, 8)

        # Promote inputs to avoid saturation, then align them
        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
        self.input_shift(shift)

        # We need a custom padding for specifics padding values
        if self._padding_value is not None:
            # Set the same scale factor as inputs
            padding_value, shift = self._padding_value.align(inputs)
            self.pad_shift(shift)

            # Note that we use the custom padding even when padding value is 0.
            inputs_padded = apply_padding(inputs.values,
                                          list(self.strides),
                                          (kernel.shape[1], kernel.shape[0]),
                                          padding_value.values)
            inputs = FixedPoint(inputs_padded, inputs.frac_bits, inputs.value_bits)

        if isinstance(self.padding, str):
            tf_padding = self.padding.upper()
        else:
            tf_padding = self.padding

        outputs = tf.nn.convolution(inputs,
                                    kernel,
                                    strides=list(self.strides),
                                    padding=tf_padding,
                                    dilations=list(self.dilation_rate),
                                    data_format=self._tf_data_format,
                                    name=self.__class__.__name__)

        if self.use_bias:
            bias = self.bias_quantizer(self.bias, training)

            # Align intermediate outputs and biases before adding them
            outputs, shift = outputs.align(bias)
            self.conv_shift(shift)
            bias, shift = bias.promote(self.buffer_bitwidth).align(outputs)
            self.bias_shift(shift)
            outputs = tf.add(outputs, bias)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        if self._padding_value is not None:
            config["padding_value"] = self._padding_value.to_float().numpy()
        return config


@tf.keras.utils.register_keras_serializable()
class QuantizedConv2DTranspose(Calibrable, layers.Conv2DTranspose):
    """A transposed convolutional layer that operates on quantized inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.weight_quantizer = deserialize_quantizer(self.quant_config, "weight_quantizer",
                                                      WeightQuantizer, True)
        if self.use_bias:
            self.bias_quantizer = deserialize_quantizer(self.quant_config, "bias_quantizer",
                                                        WeightQuantizer, True)
        self.buffer_bitwidth = self.quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        if self.use_bias:
            self.conv_shift = CalibrableVariable()
            self.bias_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # Raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError("QuantizedConv2DTranspose only accepts FixedPoint inputs. "
                            f"Receives {type(inputs)} inputs.")

        # Quantize the weights
        # Conv2DTranspose weight shape is (h, w, filters, channels) which cannot be quantized
        # 'per-axis' as it is: the last dimension is not the output last dimension (i.e filters)
        # so it must be transposed to (h, w, channels, filters) before quantization to have the
        # appropriate frac_bits shape. Kernel values will then be transposed back to apply the
        # Tensorflow operation.
        kernel = self.weight_quantizer(tf.transpose(self.kernel, (0, 1, 3, 2), training))

        # Promote inputs to avoid saturation, then align them
        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
        self.input_shift(shift)

        # Prepare deconvolution output shape
        inputs_shape = tf.shape(inputs)
        batch_size, height, width = inputs_shape[0], inputs_shape[1], inputs_shape[2]
        kernel_h, kernel_w = self.kernel_size
        stride_h, stride_w = self.strides

        if self.output_padding is None:
            out_pad_h = out_pad_w = None
        else:
            out_pad_h, out_pad_w = self.output_padding

        # Infer the dynamic output shape
        out_height = conv_utils.deconv_output_length(height,
                                                     kernel_h,
                                                     padding=self.padding,
                                                     output_padding=out_pad_h,
                                                     stride=stride_h,
                                                     dilation=self.dilation_rate[0])
        out_width = conv_utils.deconv_output_length(width,
                                                    kernel_w,
                                                    padding=self.padding,
                                                    output_padding=out_pad_w,
                                                    stride=stride_w,
                                                    dilation=self.dilation_rate[1])

        output_shape = (batch_size, out_height, out_width, self.filters)
        output_shape_tensor = tf.stack(output_shape)

        # Transpose back values so that they have the shape expected by the tensorflow op (h, w,
        # filters, channels)
        kernel_values = tf.transpose(kernel.values, (0, 1, 3, 2))

        # Do convolution on values
        conv_values = backend.conv2d_transpose(inputs.values, kernel_values, output_shape_tensor,
                                               self.strides, self.padding, self.data_format,
                                               self.dilation_rate)

        if isinstance(kernel, QFloat):
            # Multiply by the scale
            conv_values *= kernel.scales.values

        # Build a new FixedPoint
        if isinstance(kernel, FixedPoint):
            filters_frac_bits = kernel.frac_bits
        else:
            filters_frac_bits = kernel.scales.frac_bits
        outputs = FixedPoint(conv_values, inputs.frac_bits + filters_frac_bits, inputs.value_bits)

        if self.use_bias:
            # Quantize biases
            bias = self.bias_quantizer(self.bias, training)
            # Align intermediate outputs and biases before adding them
            outputs, shift = outputs.align(bias)
            self.conv_shift(shift)
            bias, shift = bias.promote(self.buffer_bitwidth).align(outputs)
            self.bias_shift(shift)
            outputs = tf.add(outputs, bias)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
BatchNormalization transformations on models.
"""
import numpy as np
import tensorflow as tf

from keras.layers import (BatchNormalization, MaxPool2D, GlobalAvgPool2D, Conv2D, SeparableConv2D,
                          Dense, DepthwiseConv2D, Conv2DTranspose)

from ..utils import deep_clone_model
from ...layers import (PaddedConv2D, QuantizedConv2D, QuantizedSeparableConv2D, QuantizedDense,
                       QuantizedDepthwiseConv2D, QuantizedConv2DTranspose, SeparableConv2DTranspose,
                       QuantizedSeparableConv2DTranspose)


def invert_batchnorm_pooling(model):
    """ Inverts pooling and BatchNormalization layers in a model to have BN layer before pooling.

    Returns a new model where pooling and batch normalization layers are inverted. From a Keras
    model where pooling layers precede batch normalization layers, this function places the BN
    layers before pooling layers. This is the first step before folding BN layers into processing
    layers.

    Note:
        Inversion of layers is equivalent only if the gammas of BN layers are positive. The
        function raises an error if not.

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: the updated model

    Raises:
        RuntimeError: if a candidate BatchNormalization layer has gamma values that are not strictly
            positive.
    """

    # Maps between successive pooling->BN layers. These pairs will be inverted when cloning.
    pool2bn_map = {}
    bn2pool_map = {}

    # Map BatchNormalization layers that have only one inbound layer being a MaxPool2D or GAP2D
    for layer in model.layers:
        if (isinstance(layer, BatchNormalization) and len(layer.inbound_nodes) == 1
                and isinstance(layer.inbound_nodes[0].inbound_layers,
                               (MaxPool2D, GlobalAvgPool2D))):
            gammas = layer.get_weights()[0]
            if isinstance(layer.inbound_nodes[0].inbound_layers, MaxPool2D) and np.any(gammas <= 0):
                # It is impossible to invert MaxPool->BN with gammas <= 0
                raise RuntimeError(f"There are {np.sum(gammas <= 0)} negative gammas in the "
                                   f"BatchNormalization layer {layer.name}. Negative gammas are "
                                   "not supported.")
            bn2pool_map[layer] = layer.inbound_nodes[0].inbound_layers
            pool2bn_map[layer.inbound_nodes[0].inbound_layers] = layer

    if not pool2bn_map:
        return model

    def replace_layer(layer):
        if layer in pool2bn_map:
            # Replace pooling layer with the corresponding BN layer
            layer_bn = pool2bn_map[layer]
            config_bn = layer_bn.get_config()
            if isinstance(layer, GlobalAvgPool2D):
                config_bn['axis'] = [-1]
            return layer_bn.from_config(config_bn)
        if layer in bn2pool_map:
            # Replace BN layer with the corresponding pooling layer
            layer_pool = bn2pool_map[layer]
            return layer_pool.from_config(layer_pool.get_config())
        return layer.from_config(layer.get_config())

    return deep_clone_model(model, clone_function=replace_layer)


def fold_batchnorms(model):
    """Returns a new model where BatchNormalization layers are folded into previous layers.

    From a Keras model where BN layers follow processing layers, this function removes the BN layers
    and updates the preceding layers weights and bias accordingly. The new model is strictly
    equivalent to the previous one.

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: the model with BN folded
    """
    # Clone model in order to modify before folding
    model_copy = deep_clone_model(model)

    # Get BN layers to fold, mapping with the preceding layer and the following layer if present
    map_prev_layer_to_bn, map_bn_to_next_layer = _find_batchnorms_to_fold(model_copy)

    # Update model before cloning: remove BN layers, update inbound nodes and output layers
    _prepare_model_to_fold_BN(model_copy, map_prev_layer_to_bn, map_bn_to_next_layer)

    # Clone model (BN layers will disappear)
    def replace_layer(layer):
        config = layer.get_config()
        # Set use_bias=True in layers where BN are folded to accept new computed weights
        if layer in map_prev_layer_to_bn:
            config['use_bias'] = True
        return layer.from_config(config)

    model_folded = deep_clone_model(model_copy, clone_function=replace_layer)

    # Set weights in the new model
    for layer in model_copy.layers:
        if layer in map_prev_layer_to_bn:
            layer_bn = map_prev_layer_to_bn[layer]
            new_weights = _compute_BN_folded_weights(layer, layer_bn)
            model_folded.get_layer(layer.name).set_weights(new_weights)

    return model_folded


def _find_batchnorms_to_fold(model):
    """ Retrieves BatchNormalization layers that can be folded.

    This is limited to BN layers that follow supported layer types. Moreover, only BN layers with
    one inbound/outbound node are folded.

    Args:
        model (keras.Model): a model

    Returns:
        dict, dict: map between a layer and its BN layer to be folded and map between the BN layer
            and its following layer (it if exists)

    Raises:
        RuntimeError: if the candidate BN has null gammas or BN is not applied on the last axis.
    """

    supported_layers = (Conv2D, PaddedConv2D, Conv2DTranspose, SeparableConv2D, DepthwiseConv2D,
                        Dense, QuantizedConv2D, QuantizedConv2DTranspose, QuantizedSeparableConv2D,
                        QuantizedDepthwiseConv2D, QuantizedDense, SeparableConv2DTranspose,
                        QuantizedSeparableConv2DTranspose)

    # Map between a layer and its following BN layer and map between the BN layer and its next layer
    map_prev_layer_to_bn = {}
    map_bn_to_next_layer = {}

    # Find triplet "layer -> BN -> next layer"
    for layer in model.layers:
        # Find a supported layer followed by a BN layer
        if (type(layer) in supported_layers and len(layer.outbound_nodes) == 1 and
                isinstance(layer.outbound_nodes[0].layer, BatchNormalization)):
            layer_bn = layer.outbound_nodes[0].layer
            # BN layer must have only one inbound node and no more than one outbound node
            bn_outbounds = layer_bn.outbound_nodes
            if (len(layer_bn.inbound_nodes) != 1 or len(bn_outbounds) > 1):
                continue

            # Null gammas are not supported: once folded, new kernel would be zero
            gamma = layer_bn.get_weights()[0]
            if np.any(gamma == 0):
                raise RuntimeError(f"There are {np.sum(gamma == 0)} null gammas in the "
                                   f"BatchNormalization layer '{layer_bn.name}'. Null gammas are "
                                   "not supported.")

            # Check BN axis parameter
            if (len(layer_bn.axis) != 1 or layer_bn.axis[0] != len(layer_bn.input_shape) - 1):
                raise RuntimeError(f"The BatchNormalization must be applied on the last "
                                   f"axis. '{layer_bn.name}' layer has axis={layer_bn.axis}.")

            map_prev_layer_to_bn[layer] = layer_bn
            if len(bn_outbounds) == 1:
                map_bn_to_next_layer[layer_bn] = bn_outbounds[0].layer

    return map_prev_layer_to_bn, map_bn_to_next_layer


def _prepare_model_to_fold_BN(model, map_prev_layer_to_bn, map_bn_to_next_layer):
    """ Prepares model for folding by removing BatchNormalization layers.

    To fold BN layers by using "deep_clone_model", the model must be prepared by modifying some
    internal variables, such as _self_tracked_trackables, _output_layers or _inbound_nodes, which
    results in by-passing the BN.

    Three operations are done here:

        1. Remove BN layers from model.layers
        2. Bypass BN layers in the graph by updating inbound nodes of the layers following the BN
           layers. The new inbound nodes are the layers preceding the BN layers
        3. If a BN layer is an output layer of the model, the preceding layer must be added to the
           new output layers

    The model and the layers are directly modified in this function.

    Args:
        model (keras.model): the model to update
        map_prev_layer_to_bn (dict): map between a layer and its BN layer to be folded
        map_bn_to_next_layer (dict): map between the BN layer and its following layer (it if exists)
    """
    # Remove BN layers from model.layers
    for layer_bn in map_prev_layer_to_bn.values():
        model._self_tracked_trackables.remove(layer_bn)

    # Update inbound nodes as there is no BN between target layer and following layer. Inbound nodes
    # are used in "deep_clone_layer" function to create the model architecture. Here, we replace the
    # inbound node of the next layer with the inbound node of the BN layer, in order to bypass the
    # BN layer.
    for layer_bn, next_layer in map_bn_to_next_layer.items():
        node = layer_bn.outbound_nodes[0]
        node_index = next_layer._inbound_nodes.index(node)
        next_layer._inbound_nodes[node_index] = layer_bn.inbound_nodes[0]

    # If BN layer is an output layer, replace it with its inbound layer
    for prev_layer, layer_bn in map_prev_layer_to_bn.items():
        if layer_bn in model._output_layers:
            # Replace BN layer in _output_layers and _output_coordinates
            index_bn = model._output_layers.index(layer_bn)
            model._output_layers[index_bn] = prev_layer
            model._output_coordinates[index_bn] = (prev_layer, 0, 0)


def _compute_BN_folded_weights(target_layer, bn_layer):
    """ Computes the new weights of a layer after folding BatchNormalization into it.

    Args:
        target_layer (keras.Layer): the layer where BN will be folded
        bn_layer (keras.Layer): the BatchNormalization layer to fold into the preceding layer

    Returns:
        list: the new weights to set in the folded destination layer.
    """
    # Get kernel and bias weights of the target layer
    kernel_position = 0
    bias_position = 1
    if isinstance(target_layer, SeparableConv2DTranspose):
        # Pointwise kernel and bias are last in the list, the first weights are the C depthwise
        # kernels
        kernel_position = -2 if target_layer.use_bias else -1
        bias_position = -1
    elif isinstance(target_layer, SeparableConv2D):
        kernel_position = 1
        bias_position = 2

    weights = target_layer.get_weights()
    kernel = weights[kernel_position]
    bias = weights[bias_position] if target_layer.use_bias else 0

    # Get BN weights
    gamma, beta, mean, var = bn_layer.get_weights()
    scale_BN = gamma / np.sqrt(var + bn_layer.epsilon)

    # Compute new folded kernel and bias
    if isinstance(target_layer, DepthwiseConv2D):
        # Last dimension of DepthwiseConv2D kernel is 1 so scale_BN should be reshaped to (shape, 1)
        # (first two dimension will be broadcasted)
        new_kernel = kernel * scale_BN.reshape(*scale_BN.shape, 1)
    elif isinstance(target_layer, Conv2DTranspose):
        # Handle Conv2DTranspose kernels that are (h, w, filters, channels) instead of (h, w,
        # channels, filters)
        new_kernel = tf.transpose(kernel, (0, 1, 3, 2)) * scale_BN
        new_kernel = tf.transpose(new_kernel, (0, 1, 3, 2))
    else:
        new_kernel = kernel * scale_BN
    new_bias = beta + (bias - mean) * scale_BN

    # Return all weights with modified ones
    new_weights = weights
    new_weights[kernel_position] = new_kernel
    if target_layer.use_bias:
        new_weights[bias_position] = new_bias
    else:
        new_weights.append(new_bias)

    return new_weights



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from ...debugging import assert_equal
from ..fixed_point import FixedPoint


@tf.experimental.dispatch_for_api(tf.add)
def fp_add(x: FixedPoint, y: FixedPoint):
    return x + y


@tf.experimental.dispatch_for_api(tf.clip_by_value)
def fp_clip_by_value(t: FixedPoint,
                     clip_value_min: FixedPoint,
                     clip_value_max: FixedPoint,
                     name=None):
    """Clips tensor values to a specified min and max.

    Args:
        t (:obj:`FixedPoint`): the FixedPoint to be clipped.
        clip_value_min (:obj:`FixedPoint`): the minimum value.
        clip_value_max (:obj:`FixedPoint`): the maximum value.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the clipped FixedPoint.
    """
    # The fractional bits of all inputs must be equal
    assert_equal(t.frac_bits, clip_value_min.frac_bits)
    assert_equal(t.frac_bits, clip_value_max.frac_bits)
    clip_values = tf.clip_by_value(t.values, clip_value_min.values, clip_value_max.values, name)
    return FixedPoint(clip_values, t.frac_bits, t.value_bits)


@tf.experimental.dispatch_for_api(tf.math.reduce_sum)
def fp_reduce_sum(input_tensor: FixedPoint, axis=None, keepdims=False, name=None):
    """Computes the sum of elements across dimensions of a FixedPoint.

    Args:
        input_tensor (:obj:`FixedPoint`): the FixedPoint to be summed.
        axis (list, optional): the dimensions to reduce. If None, reduces all
            dimensions. Defaults to None.
        keepdims (bool, optional): if true, retains reduced dimensions with length 1.
            Defaults to False.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the summed FixedPoint.
    """
    # input_values must be quantized per-tensor
    input_tensor.assert_per_tensor()

    # Reduce sum
    s_values = tf.math.reduce_sum(
        input_tensor.values, axis, keepdims=keepdims, name=name)

    # Return a new FixedPoint
    return FixedPoint(s_values, input_tensor.frac_bits, input_tensor.value_bits)


@tf.experimental.dispatch_for_api(tf.math.reduce_max)
def fp_reduce_max(input_tensor: FixedPoint,
                  axis=None,
                  keepdims=False,
                  name=None):
    """Computes the maximum of elements across dimensions of a FixedPoint.

    Args:
        input_tensor (:obj:`FixedPoint`): the FixedPoint to be estimated.
        axis (list, optional): the dimensions to reduce. If None, reduces all
            dimensions. Defaults to None.
        keepdims (bool, optional): if true, retains reduced dimensions with length 1.
            Defaults to False.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the maximum FixedPoint.
    """
    # We only support reduce_max if the input is per-tensor
    input_tensor.assert_per_tensor()

    # Reduce max
    s_values = tf.math.reduce_max(input_tensor.values,
                                  axis,
                                  keepdims=keepdims,
                                  name=name)
    # Return a new FixedPoint
    return FixedPoint(s_values, input_tensor.frac_bits, input_tensor.value_bits)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import warnings
import tensorflow as tf

from ..fixed_point import FixedPoint


def warn_float_conversion(x, op):
    if tf.executing_eagerly():
        x_name = "FixedPoint"
    else:
        x_name = x.name
    warnings.warn(f"Reverting {x_name} to float to perform {op} with Tensor")


@tf.experimental.dispatch_for_binary_elementwise_apis(FixedPoint, tf.Tensor)
def fp_tensor_binary_elementwise_api_handler(api_func, x, y):
    warn_float_conversion(x, api_func)
    return api_func(x.to_float(), y)


@tf.experimental.dispatch_for_binary_elementwise_apis(tf.Tensor, FixedPoint)
def tensor_fp_binary_elementwise_api_handler(api_func, x, y):
    warn_float_conversion(y, api_func)
    return api_func(x, y.to_float())



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from .qtensor import QTensor, round_through, ceil_through, pow2, ceil_log2
from .fixed_point import FixedPoint


class QFloat(QTensor):
    """A Tensor of integer values representing quantized float numbers

    Args:
        values (:obj:`tensorflow.Tensor`): a tensor of integer values
        scales (:obj:`tensorflow.Tensor`): a FixedPoint of scales
        value_bits (int): the number of value bits
    """
    scales: FixedPoint = FixedPoint(1.0, 0, 32)

    def __init__(self, values, scales, value_bits):
        self.scales = scales
        # We store integer values in a float tensor
        values = tf.convert_to_tensor(values, dtype=tf.float32)
        self.values = QTensor.clamp(values, value_bits)
        self.shape = self.values.shape
        self.value_bits = value_bits

    @property
    def per_tensor(self):
        return tf.rank(self.scales.values) == 0

    @staticmethod
    def quantize(x, float_max, value_bits, scales_bits):
        r"""Converts a float Tensor to a QFloat

        It first evaluates the maximum integer values that can be stored for
        the specified value bits: int_max = 2^bits - 1.

        It then converts the original float values into integer values so that:

        .. math:: {x_{int}} = round(x * \\frac{int\_max}{float\_max})

        The resulting integer values are clipped to [-int_max-1, int_max].

        Args:
            x (:obj:`tensorflow.Tensor`): a tensor of float values.
            float_max (:obj:`tensorflow.Tensor`): a tensor of maximum values
            value_bits (int): the number of value bits
            scales_bits (int): the number of scales bits

        Returns:
            :obj:`QFloat`: the QFloat representation.
        """
        # Evaluate QFloat scales
        scales = tf.cast(float_max, tf.float32) / QTensor.int_max(value_bits)
        scales_int_bits = ceil_log2(scales)
        scales_frac_bits = scales_bits - scales_int_bits
        # We do not apply the default FixedPoint quantization algorithm, as it
        # uses a round operation, and we want to make sure the quantized scales
        # are at least equal to the target scales. Should they be lower, this
        # would lead to a reduction of the QFloat quantization range and values
        # near float_max would be clipped.
        scales_values = ceil_through(scales * pow2(scales_frac_bits))
        fp_scales = FixedPoint(scales_values, scales_frac_bits, scales_bits)
        # Quantize the inputs using the approximated FixedPoint scales
        values = round_through(x / fp_scales.to_float())
        return QFloat(values, fp_scales, value_bits)

    def to_float(self):
        """Returns a float representation of the QFloat

        Returns:
            :obj:`tensorflow.Tensor`: the float representation.
        """
        return self.values * self.scales.to_float()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
quantizeml main command-line interface.
"""

import argparse
import json
import os
import sys

from .models import (load_model, quantize, dump_config, save_weights, load_weights,
                     check_quantization)
from .models.transforms import fold_rescaling


def quantize_model(model_path, quant_config, fold=True, name=False, add_deq=True):
    """ CLI entry point to quantize a model using the provided configuration.

    Args:
        model_path (str): Path to the model to quantize.
        quant_config (str): Path to the quantization configuration file.
        fold (bool, optional): Whether to collapse foldable layers before quantizing.
            Defaults to True.
        name (bool, optional): Whether to print the quantized output model filename.
            Defaults to False.
        add_deq (bool, optional): allows to convert output to float. Defaults to True.
    """
    # Build name for the output model
    model_name = os.path.splitext(model_path)[0]
    config_name = os.path.splitext(os.path.basename(quant_config))[0]
    output_name = f"{model_name}_{config_name}.h5"

    # When arg.name is requested, simply print out the output model name
    if name:
        print(output_name)
        exit(0)

    # Load the configuration file and the model
    with open(quant_config) as f:
        config = json.load(f)
    model = load_model(model_path)

    # Fold layers
    if fold:
        print(f"Collapsing foldable layers of {model_path}.")
        model = fold_rescaling(model)

    # Quantize the model and save it
    print(f"Quantizing model {model_path} with configuration file {quant_config}.")
    model_q = quantize(model, config, add_deq)
    model_q.save(output_name, include_optimizer=False)
    print(f"Saved quantized model to {output_name}.")


def dump_model_config(model_path, skip_default=False, output_name=None):
    """ CLI entry point to dump the quantization configuration from a model.

    Args:
        model_path (str): Path to the model to extract the configuration from.
        skip_default (bool): Whether to skip default values on each quantizer. Defaults to False.
        output_name (str): Path to save the configuration.
            Defaults to <model_path>_quant_config.json.
    """
    # Build name for the output model
    if output_name is None:
        model_name = os.path.splitext(model_path)[0]
        output_name = f"{model_name}_quant_config.json"

    # Load the model and get its quantization configuration
    model = load_model(model_path)
    config = dump_config(model, skip_default=skip_default)
    with open(output_name, "w") as f:
        json.dump(config, f, indent=4)
    print(f"Saved quantization configuration to {output_name}.")


def save_model_weights(model_path, weights_path):
    """ CLI entry point to save the model weights in an npz file.

    Args:
        model_path (str): Path to the model to extract the weights from.
        weights_path (str): Path to save the npz file.
            Defaults to <model_path>.npz.
    """
    # Build name for weights file
    if weights_path is None:
        model_name = os.path.splitext(model_path)[0]
        weights_path = f"{model_name}.npz"

    # Load the model and save its weights
    model = load_model(model_path)
    save_weights(model, weights_path)
    print(f"Saved model weights to {weights_path}.")


def load_model_weights(model_path, weights_path):
    """ CLI entry point to apply weights to a model from an npz file.

    Args:
        model_path (str): Path to the model on which apply the weights.
        weights_path (str): Path to load the npz file.
    """
    # Update the model weights with the npz file
    model = load_model(model_path)
    load_weights(model, weights_path)
    model.save(model_path, include_optimizer=False)
    print(f"Saved model with new weights to {model_path}.")


def check_quantized_model(model_path):
    """ CLI entry point to check model quantization.

    Args:
        model_path (str): Path to the model on which apply the weights.
    """
    model = load_model(model_path)
    messages = check_quantization(model)
    for msg in messages:
        print(msg)


def main():
    """ CLI entry point.

    Contains an argument parser with specific arguments depending on the model to be created.
    Complete arguments lists available using the -h or --help argument.

    """
    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers(dest="action")
    sp.add_parser("version", help="Display quantizeml version.")

    # Quantize arguments
    q_parser = sp.add_parser(
        "quantize", help="Quantize an input model, given a quantization configuration file.")
    q_parser.add_argument("-m", "--model", type=str, required=True, help="Model to quantize")
    q_parser.add_argument("-c", "--quantization_config", type=str,
                          required=True, help="Quantization configuration file")
    q_parser.add_argument("-n", "--name", action="store_true",
                          help="Print quantized output model filename")
    q_parser.add_argument("-f", "--fold", action="store_true",
                          help="Collapse foldable layers before quantizing")
    q_parser.add_argument("-nd", "--no_deq", action="store_false",
                          help="Do not add a dequantizer after head")

    # Dump config arguments
    c_parser = sp.add_parser("config", help="Extract quantization configuration from a model.")
    c_parser.add_argument("-m", "--model", type=str, required=True,
                          help="Model to extract config from.")
    c_parser.add_argument("-sd", "--skip_default", action="store_true",
                          help="Remove default values on each quantizer. Defaults to %(default)s.")
    c_parser.add_argument("-o", "--output_path", type=str, help="Store quantization configuration. "
                          "Defaults to <model>_quant_config.json")

    # Save weights arguments
    w_parser = sp.add_parser("save_weights", help="Store models weights in an npz.")
    w_parser.add_argument("-m", "--model", type=str, required=True,
                          help="Model to extract weights from.")
    w_parser.add_argument("-w", "--weights_path", type=str,
                          help="Npz file that contains the saved weights.")

    # Load weights arguments
    w_parser = sp.add_parser("load_weights", help="Apply weights to a model from an npz file.")
    w_parser.add_argument("-m", "--model", type=str, required=True,
                          help="Model to apply the weights.")
    w_parser.add_argument("-w", "--weights_path", type=str, required=True,
                          help="Npz file that contains the weights to apply.")

    # Check action and arguments
    c_parser = sp.add_parser(
        "check", help="Check the quantization of a model to detect issues that can be fixed "
        "by adjusting the quantization configuration.")
    c_parser.add_argument("-m", "--model", type=str, required=True, help="Model to check")

    args = parser.parse_args()

    if args.action == "version":
        # importlib.metadata was introduced in Python 3.8 and is available to older versions as the
        # importlib-metadata project
        if sys.version_info >= (3, 8):
            from importlib import metadata
        else:
            import importlib_metadata as metadata
        print(metadata.version('quantizeml'))
    elif args.action == "quantize":
        quantize_model(
            model_path=args.model,
            quant_config=args.quantization_config,
            fold=args.fold,
            name=args.name,
            add_deq=args.no_deq
        )
    elif args.action == "config":
        dump_model_config(
            model_path=args.model,
            skip_default=args.skip_default,
            output_name=args.output_path,
        )
    elif args.action == "save_weights":
        save_model_weights(
            model_path=args.model,
            weights_path=args.weights_path,
        )
    elif args.action == "load_weights":
        load_model_weights(
            model_path=args.model,
            weights_path=args.weights_path,
        )
    elif args.action == "check":
        check_quantized_model(args.model)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
import keras

from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH
from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer

__all__ = ["QuantizedDense"]


@tf.keras.utils.register_keras_serializable()
class QuantizedDense(Calibrable, keras.layers.Dense):
    """A Dense layer that operates on quantized inputs and weights

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.weight_quantizer = deserialize_quantizer(
            self.quant_config, "weight_quantizer", WeightQuantizer, True)
        if self.use_bias:
            self.bias_quantizer = deserialize_quantizer(
                self.quant_config, "bias_quantizer", WeightQuantizer, True)
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        if self.use_bias:
            self.prod_shift = CalibrableVariable()
            self.bias_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint or tf.Tensor
        if not isinstance(inputs, (FixedPoint, tf.Tensor)):
            raise TypeError(f"QuantizedDense only accepts FixedPoint or tf.Tensor\
                              inputs. Receives {type(inputs)} inputs.")

        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel, training)

        if isinstance(inputs, tf.Tensor):
            # Assume the inputs are integer stored as float, which is the only tf.Tensor
            # inputs that are allowed
            inputs = FixedPoint(inputs, 0, 8)

        # Prepare inputs before invoking matmul:
        # - promote to a higher bitwidth to avoid overflows,
        # - align all channels on the same fractional bits
        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
        # - update shift values if calibration is enabled
        self.input_shift(shift)

        outputs = tf.matmul(inputs, kernel)

        if self.use_bias:
            # Quantize biases
            bias = self.bias_quantizer(self.bias, training)
            # Promote them before alignment to avoid overflow
            bias = bias.promote(self.buffer_bitwidth)
            # Align intermediate outputs and biases before adding them
            outputs, shift = outputs.align(bias)
            self.prod_shift(shift)
            bias, shift = bias.align(outputs)
            self.bias_shift(shift)
            outputs = tf.add(outputs, bias)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np
import tensorflow as tf
import keras

from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, OutputQuantizer
from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH


__all__ = ["QuantizedGELU", "QuantizedReLU"]


@tf.keras.utils.register_keras_serializable()
class QuantizedGELU(Calibrable, keras.layers.Layer):
    """Quantize the `Gaussian Error Linear Units (GELU) <https://arxiv.org/abs/1606.08415>`_
    activation, following the implementation on `i-BERT <https://arxiv.org/abs/2101.01321>`_.

    To quantize non-linear activation functions, three ways are commonly used:

        1. Dequantize the input, make the activation function in float and then quantize the output.
        2. Store the quantized input in a lookup table, and use it to quantize.
        3. Approximate the activation function with a linear function.

    Follow the explanation given on `i-BERT <https://arxiv.org/abs/2101.01321>`_, in order
    to reduce on-chip memory cost and increase efficient hardware performance, the third way
    is used here.

    The GELU activation function is defined as:

    .. math::
        GELU(x) = 0.5*x*[1 + erf(\\frac{x}{sqrt(2)})]

    Where erf(:math:`x`) is the error function. `i-BERT <https://arxiv.org/abs/2101.01321>`_ shows
    that a degree two polynomial approximation to the erf function would be given by

    .. math::
        L(x) = sgn(x)[a(clip(|x|, max = -b) + b)^2 + 1]

    With :math:`a = -0.2888`, :math:`b = -1.769`, and average error / maximum error of
    :math:`8.2\\times10^{-3} / 1.8\\times10^{-2}` (respectively), within the range :math:`|x| < 1`.

    Args:
        quant_config (dict): quantization configuration of input/output quantizers.
    """
    unsupported_args = {
        'approximate': True}

    def __init__(self, *args, name="i_gelu", quant_config={}, **kwargs):
        super().__init__(*args, name=name, **kwargs)
        self.quant_config = quant_config
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth >= 0, "The buffer_bitwidth must be a non-negative integer."
        assert self.buffer_bitwidth >= 23, "The buffer_bitwidth must be >= 23."
        # Coefficients of the degree 2 polynomial, used to approximate erf(x/sqrt(2))
        self._a_int = FixedPoint.quantize(-0.2888, self.buffer_bitwidth, 11)
        self._b_int = FixedPoint.quantize(-1.769, self.buffer_bitwidth, 16)
        self._c_int = FixedPoint.quantize(1.0, self.buffer_bitwidth, 0)
        # Coefficients used in i_gelu function
        self._s2_int = FixedPoint.quantize(1.0 / 1.4142, self.buffer_bitwidth, 6)
        self.intermediate_quantizer = deserialize_quantizer(
            self.quant_config, "intermediate_quantizer", OutputQuantizer, True)
        assert self.quant_config["intermediate_quantizer"].get("axis") != "per-axis", \
            "The intermediate quantizer cannot be per-axis."
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)

        # Add objects that will store the shift values.
        self.q_in_shift = CalibrableVariable()
        self.q_b_shift = CalibrableVariable()
        self.q_inter_shift = CalibrableVariable()
        self.q_c_shift = CalibrableVariable()
        self.q_erf_shift = CalibrableVariable()

    @property
    def _pol_coeff(self):
        """Return the coefficients of the degree 2 polynomial approximation to the erf function."""
        return self._a_int, self._b_int, self._c_int

    def i_poly(self, q_in, q_a, q_b, q_c):
        """Integer-only Computation of Second-order Polynomial :math:`a*(x + b)^2 + c`

        Args:
            q_in (:obj:`FixedPoint`): an integer quantized input tensor.
            q_a (obj:`FixedPoint`): first coefficient of the polynomial, :math:`q_a = S*a`.
            q_b (obj:`FixedPoint`): second coefficient of the polynomial, :math:`q_b = S*b`.
            q_c (obj:`FixedPoint`): third coefficient of the polynomial, :math:`q_c = S*c`.

        Returns:
            :obj:`FixedPoint`: the y-tensor: :math:`y \approx a(x + b)^2 + c`

        Note:
            ``S`` denotes the quantization scale.
        """
        # Add intermediate inputs
        q_sum = q_in + q_b
        # Downscale to reduce the frac_bits before squaring the q_sum
        q_sum = self.intermediate_quantizer(q_sum)
        q_sum = q_sum.promote(self.buffer_bitwidth)
        q_inter = q_sum**2
        # Downscale to reduce the frac_bits before multiply by q_a
        q_inter, shift = q_inter.downscale(self.buffer_bitwidth // 2, self.buffer_bitwidth)
        self.q_inter_shift(shift)
        q_inter = q_inter * q_a

        # Align intermediate results before adding them
        q_c, shift = q_c.align(q_inter)
        self.q_c_shift(shift)
        return q_inter + q_c

    def i_erf(self, q_in):
        """Approximate the error function with a degree two polynomial.

        Args:
            q_in (:obj:`FixedPoint`): an integer quantized input tensor.

        Returns:
            :obj:`FixedPoint`: the y-tensor: :math:`y \approx i_erf(x)`
        """
        # Get the coefficients of the degree 2 polynomial
        q_a, q_b, q_c = self._pol_coeff
        # Align intermediate inputs before clipping them
        q_in, shift = q_in.align(q_b)
        self.q_in_shift(shift)
        q_b, shift = q_b.align(q_in)
        self.q_b_shift(shift)
        # Express zero as a FixedPoint with the same frac_bits as q_in because this
        # is what clip_by_value expects. The actual hardware implementation will
        # simply use a zero integer.
        zero = FixedPoint(0, q_in.frac_bits, q_in.value_bits)
        # Compute erf function, as a degree 2 polynomial approximation
        q_in_abs = tf.clip_by_value(q_in.abs(), zero, q_b.abs())
        return self.i_poly(q_in_abs, q_a, q_b, q_c) * q_in.sign

    def i_gelu(self, q_in):
        """Approximate the GELU activation function with a degree two polynomial.

        Args:
            q_in (:obj:`FixedPoint`): an integer quantized input tensor.

        Returns:
            :obj:`FixedPoint`: the y-tensor: :math:`y \approx i_gelu(x)`
        """
        # Compute i_erf function, the result is between -1 and 1
        q_erf_input = q_in.promote(self.buffer_bitwidth) * self._s2_int
        q_erf = self.i_erf(q_erf_input)
        # Downscale before multiplying to avoid saturation
        erf_value_bits = self.buffer_bitwidth - q_in.value_bits - 1
        q_erf, shift = q_erf.downscale(erf_value_bits, self.buffer_bitwidth)
        self.q_erf_shift(shift)
        # Compute i-GELU function
        # The value (q_erf + 1) should be multiplied by 0.5: a right shift is done instead
        return ((q_erf + 1) >> 1) * q_in

    def call(self, inputs, training=None):
        """i-GELU forward in three steps:
            1. Quantize the inputs (if needed)
            2. Pass inputs through the i-GELU function
            3. Update the new quantization scale

        Args:
            inputs (:obj:`FixedPoint`): the inputs tensor.
            training (bool, optional): the training mode. Defaults to None.

        Returns:
            :obj:`FixedPoint`: float of i-GELU(inputs).
        """
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedGeLU only accepts FixedPoint inputs.\
                             Receives {type(inputs)} inputs.")

        # Approximate the GELU function with i-GELU
        outputs = self.i_gelu(inputs)

        # 3. Update the new quantization scale in output quantizer
        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training=training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config


@tf.keras.utils.register_keras_serializable()
class QuantizedReLU(Calibrable, keras.layers.Layer):
    """Quantized version of the ReLU activation layer applicable on FixedPoint tensor.
    """
    unsupported_args = {
        'negative_slope': 0,
        'threshold': 0}

    def __init__(self, *args, max_value=6, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config

        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)

        if max_value is None:
            raise ValueError(
                f"QuantizedReLU layer {self.name} has been initialized with\
                    unsupported None max_value.")
        # Store max_value
        if isinstance(max_value, np.ndarray):
            max_value = float(max_value.item())
        self.max_value = max_value

        # Also store quantized max_value as a FixedPoint
        self.qmax_value = FixedPoint.quantize(max_value, self.buffer_bitwidth)

        # We need a Calibrable to store the max_value shift (inputs dependent)
        self.max_value_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        """ReLU activation function. In other terms:
            1. clip the value between 0 and :attr:`max_value`.
            2. quantize the output if an output_quantizer is set.

        Args:
            inputs (:obj:`FixedPoint` or :obj:`tf.Tensor`): the inputs tensor.
            training (bool, optional): the training mode. Defaults to None.

        Returns:
            :obj:`FixedPoint`: QuantizedReLU outputs.
        """
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedReLU only accepts FixedPoint inputs.\
                             Receives {type(inputs)} inputs.")

        # Express zero as a FixedPoint with the same frac_bits as the inputs
        # because this is what clip_by_value expects. The actual hardware implementation
        # will simply use a zero integer.
        zero = FixedPoint(0, inputs.frac_bits, inputs.value_bits)
        # Align max_value with the inputs
        qmax_value, shift = self.qmax_value.downscale(inputs.frac_bits,
                                                      self.qmax_value.value_bits)
        self.max_value_shift(shift)
        outputs = tf.clip_by_value(inputs, zero, qmax_value)
        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training=training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({"max_value": self.max_value})
        config["quant_config"] = self.quant_config
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Rescaling transformation for quantized models.
"""
import numpy as np

from keras.layers import Rescaling, Conv2D, Dense

from quantizeml.layers.convolution import PaddedConv2D

from ...models.utils import deep_clone_model

__all__ = ["fold_rescaling"]


def fold_rescaling(model):
    """Folds the Rescaling layer of the model into the next layer weights.

    Args:
        model (keras.Model): a Keras model to fold

    Returns:
        keras.Model: the model with Rescaling layer folded
    """
    supported_dst_layers = [Conv2D, Dense]

    # Create a clone model
    clone_model = deep_clone_model(model)

    # Find the model has a Rescaling layer and check limitations on folding
    rescaling_layer = None
    dst_layer = None

    for ly in clone_model.layers:
        if isinstance(ly, Rescaling):
            # Folding is limited to single outbound node Rescaling layers
            if len(ly.outbound_nodes) != 1:
                raise ValueError("Found a Rescaling layer in the model but with multiple "
                                 "outbounds which is not supported.")
            # Retrieve the destination layer and check its type
            dst_layer = ly.outbound_nodes[0].layer
            if type(dst_layer) not in supported_dst_layers:
                raise ValueError(f"Layer type {type(dst_layer)} after Rescaling not supported, "
                                 "must be Conv2D or Dense.")
            # When destination layer is a Conv2D with padding 'same', reject:
            #   - offsets defined per axis,
            #   - scales defined per axis when offset is not null.
            # Otherwise it would require a padding value per-axis which is not supported.
            same_conv2d = (isinstance(dst_layer, Conv2D)
                           and dst_layer.get_config()['padding'].lower() == 'same')
            offset_per_axis = isinstance(ly.offset, list) and len(ly.offset) > 1
            scale_per_axis = isinstance(ly.scale, list) and len(ly.scale) > 1
            if same_conv2d and (offset_per_axis or ly.offset != 0 and scale_per_axis):
                raise NotImplementedError("Folding with offset per-axis or scale per-axis with an "
                                          "offset into a Conv2D with 'same' padding is not "
                                          "supported.")
            # Rescaling was found, exit the loop
            rescaling_layer = ly
            break

    # If no Rescaling was found, nothing to do, simply return the original model
    if rescaling_layer is None:
        return clone_model

    # Remove the Rescaling layer so that it will be ignored by clone_model
    clone_model._self_tracked_trackables.remove(rescaling_layer)

    # Bypass the Rescaling layer by connecting its inbound node to the dst_layer inbound node
    dst_layer._inbound_nodes[0] = rescaling_layer.inbound_nodes[0]

    # Clone model, rescaling layer will disappear in this step
    def replace_layer(layer):
        config = layer.get_config()
        if layer is dst_layer:
            # Force bias if there is a rescaling offset
            if rescaling_layer.offset != 0:
                config['use_bias'] = True
                # Replace Conv2D with 'same' padding by PaddedConv2D with correct padding value
                if isinstance(dst_layer, Conv2D) and config['padding'].lower() == 'same':
                    # Both offset and scale have a single value at this point
                    offset = rescaling_layer.offset[0] if isinstance(
                        rescaling_layer.offset, list) else rescaling_layer.offset
                    scale = rescaling_layer.scale[0] if isinstance(
                        rescaling_layer.scale, list) else rescaling_layer.scale
                    config['padding_value'] = -offset / scale
                    return PaddedConv2D.from_config(config)
        return layer.__class__.from_config(config)

    folded_model = deep_clone_model(clone_model, clone_function=replace_layer)
    folded_layer = folded_model.get_layer(dst_layer.name)

    # Set weights in the layer of the new model, using the Rescaling parameters:
    base_weights = dst_layer.get_weights()
    new_w = base_weights[0].copy()
    filters = new_w.shape[-1]
    # new_weights = base_weights * scale,
    for i in range(filters):
        # Rescale weights filter by filter to enable broadcast if scales are per channel
        new_w[..., i] *= rescaling_layer.scale
    new_weights = [new_w]

    if folded_layer.use_bias:
        # Build zero initialized biases if the original layer didn't have any
        new_biases = base_weights[1].copy() if dst_layer.use_bias else np.zeros(filters)
        # new_biases = base_biases + sum(base_weights_per_filter * offset)
        for i in range(filters):
            # Rescale biases filter by filter to enable broadcast if offsets are per channel
            w_i = base_weights[0][..., i]
            new_biases[i] += np.sum(w_i * rescaling_layer.offset)
        new_weights += [new_biases]

    folded_layer.set_weights(new_weights)
    return folded_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import keras
import tensorflow as tf

from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH
from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, OutputQuantizer


__all__ = ["Add", "QuantizedAdd"]


@tf.keras.utils.register_keras_serializable()
class Add(keras.layers.Layer):
    """Wrapper class of `keras.layers.Add` that allows to average inputs.
    We only support a tuple of two inputs with same shape.

    Args:
        average (bool, optional): if `True`, compute the average across all inputs.
            Defaults to `False`.
    """

    def __init__(self, *args, average=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.average = average

    def build(self, input_shape):
        if not isinstance(input_shape, (list, tuple)) or len(input_shape) != 2:
            raise ValueError(f"{self.__class__.__name__} expects two input tensors")

    def call(self, inputs, training=False):
        a, b = inputs
        output = tf.add(a, b)
        if self.average:
            output /= 2
        return output

    def get_config(self):
        config = super().get_config()
        config["average"] = self.average
        return config


@tf.keras.utils.register_keras_serializable()
class QuantizedAdd(Calibrable, Add):
    """Sums two inputs and quantize the output.

    The two inputs must be provided as a list or tuple of FixedPoint or Tensors.

    The outputs are quantized according to the specified quantization configuration.

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        # Add objects that will store the shift values.
        self.a_shift = CalibrableVariable()
        self.b_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        a, b = inputs
        if not (isinstance(a, FixedPoint) and isinstance(b, FixedPoint)):
            # If any of the inputs is not a FixedPoint, raise an error
            raise TypeError(f"QuantizedAdd only accepts FixedPoint\
                              inputs. Receives {(type(a), type(b))} inputs.")

        # Promote a before performing the addition
        a = a.promote(self.buffer_bitwidth)
        b = b.promote(self.buffer_bitwidth)

        # Align intermediate inputs before adding them
        a, shift_ab = a.align(b)
        b, shift_ba = b.align(a)

        outputs = tf.add(a, b)
        if self.average:
            outputs = outputs >> 1

        # Compute shifts
        self.a_shift(shift_ab)
        self.b_shift(shift_ba)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training=training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from .qtensor import QTensor, round_through, floor_through, pow2, ceil_log2
from ..debugging import assert_equal, assert_none_equal, assert_less, assert_less_equal


class FixedPoint(QTensor):
    """A Tensor of integer values representing fixed-point numbers

    Args:
        values (:obj:`tensorflow.Tensor`): a tensor of integer values
        frac_bits (:obj:`tensorflow.Tensor`, optional): an integer tensor of
            fractional bits. Defaults to 0.
        value_bits (int, optional): the number of value bits. Defaults to 32.
    """
    frac_bits: tf.Tensor = 0.

    def __init__(self, values, frac_bits=0, value_bits=32):
        # We store fractional bits in a float tensor to speed up calculations
        if isinstance(frac_bits, tf.Tensor):
            self.frac_bits = tf.cast(frac_bits, tf.float32)
        else:
            self.frac_bits = tf.convert_to_tensor(frac_bits, tf.float32)
        assert_less_equal(0, self.frac_bits, "The frac_bits must be positive.")
        self.value_bits = value_bits
        # We store integer values in a float tensor to speed up calculations
        if isinstance(values, tf.Tensor):
            values = tf.cast(values, tf.float32)
        else:
            values = tf.convert_to_tensor(values, dtype=tf.float32)
        # Clamp to fixed-point boundaries
        self.values = QTensor.clamp(values, value_bits)
        self.shape = self.values.shape

    @property
    def per_tensor(self):
        return tf.rank(self.frac_bits) == 0

    @staticmethod
    def quantize(x, value_bits, frac_bits=None):
        r"""Converts a float Tensor to a FixedPoint

        It converts the original float values into integer values so that:

        .. math:: {x_{int}} = round(x * 2^{frac\_bits})

        Note: 2^-frac_bits represents the FixedPoint precision.

        Before returning, the resulting integer values are clipped to the
        maximum integer values that can be stored for the specified value bits:

            [-2^value_bits, 2^value_bits - 1]

        If frac_bits is not specified, the method starts by evaluating the number
        of bits to dedicate to represent the integer part of the float tensor,
        clipped to value_bits:

        .. math:: int_bits = ceil(log2(x))

        Note: this number can be negative when x < 1.

        It then evaluates the offset of the least significant bit of the fractional
        part of the float starting from zero. This represents the fractional bits:

        .. math:: frac_bits = value_bits - int_bits

        Args:
            x (:obj:`tensorflow.Tensor`): a tensor of float values.
            value_bits (int): the number of value bits
            frac_bits (:obj:`tensorflow.Tensor`, optional): an integer tensor of fractional bits.
                Defaults to None.

        Returns:
            a :obj:`FixedPoint`
        """
        if frac_bits is None:
            if isinstance(x, int) or tf.reduce_all(x == tf.math.round(x)):
                # The input does not need to be quantized
                frac_bits = 0
            else:
                # Evaluate integer bits
                int_bits = ceil_log2(tf.abs(x))
                frac_bits = tf.maximum(0, value_bits - int_bits)
        # Project float into integer fixed-point space
        values = round_through(x * pow2(frac_bits))
        return FixedPoint(values, frac_bits, value_bits)

    @property
    def sign(self):
        """Returns the sign of the FixedPoint

        Returns:
            :obj:`FixedPoint`: the sign as a FixedPoint.
        """
        return FixedPoint(tf.math.sign(self.values), 0, self.value_bits)

    def to_float(self):
        return self.values / pow2(self.frac_bits)

    def promote(self, bits):
        """Increase the number of value bits

        Args:
            bits (int): the new number of value bits

        Returns:
            :obj:`FixedPoint`: a FixedPoint with increased value bits
        """
        if not isinstance(bits, int):
            raise TypeError("Bitwidth must be an integer")
        if bits < 0:
            raise ValueError("Invalid bitwidth")
        if bits < self.value_bits:
            raise ValueError(f"Cannot reduce value bits from {self.value_bits} to {bits}: "
                             "use a quantizer instead")
        return FixedPoint(self.values, self.frac_bits, bits)

    def align(self, other=None):
        """Align fractional bits

        This returns an equivalent FixedPoint with a scalar fractional bit
        corresponding to the maximum of:

            - the initial fractional bits on all channels,
            - when specified, the other FixedPoint fractional bits on all channels.

        This is required before performing an operation that adds or subtracts
        elements along the last dimension, to make sure all these elements are
        in the same scale.

        Args:
            other(:obj:`FixedPoint`): a FixedPoint to align to

        Returns:
            tuple(:obj:`FixedPoint`, :obj:`tf.Tensor`): a new FixedPoint with aligned
                fractional bits and the shift that was applied.
        """
        max_frac_bits = tf.reduce_max(self.frac_bits)
        if other is not None:
            if not isinstance(other, FixedPoint):
                raise ValueError("Other must be a FixedPoint")
            max_frac_bits = tf.math.maximum(max_frac_bits, tf.reduce_max(other.frac_bits))
        shift = max_frac_bits - self.frac_bits
        assert_less(shift, self.value_bits, "Cannot align the FixedPoint as it saturates its "
                    f"{self.value_bits}-bit buffer.")
        values = FixedPoint._lshift(self.values, shift)
        return FixedPoint(values, max_frac_bits, self.value_bits), shift

    def downscale(self, frac_bits, value_bits):
        """Encode a FixedPoint with a lower bitwidth

        It applies a left/right shift to the source FixedPoint values to align
        them on the target maximum fractional bits.

        It then clips them to the boundaries specified by the target value bits.

        Args:
            frac_bits (:obj:`tensorflow.Tensor`): the target fractional bits
            value_bits (int): the target value bits

        Returns:
            a :obj:`FixedPoint`
        """
        if value_bits > self.value_bits:
            raise ValueError(
                f"Cannot increase value bits from {self.value_bits} to {value_bits}:"
                "use a promotion instead")
        frac_bits = tf.cast(frac_bits, tf.float32)
        # Evaluate the shift to apply to reach the target precision
        shift = frac_bits - self.frac_bits
        assert_less(tf.abs(shift), self.value_bits, "Cannot downscale the FixedPoint as it "
                    f"saturates its {self.value_bits}-bit buffer.")
        # The shift can be positive (left-shift) or negative (right-shift)
        # In float it can be simulated by flooring after multiplying by the
        # positive/negative shift PoT:
        # - for positive shift, the results are integer and can safely be floored,
        # - for negative shift, flooring simulates a right_shift.
        values = round_through(self.values * pow2(shift))
        # return a new FixedPoint with the target precision and bitwidth
        return FixedPoint(values, frac_bits, value_bits), shift

    @staticmethod
    def _rshift(values, shift):
        return floor_through(values / pow2(shift))

    @staticmethod
    def _lshift(values, shift):
        return values * pow2(shift)

    def shift(self, s):
        """Apply a tensor-wide left or right shift.

        This takes a tensor of shift values and apply them on each item of the
        FixedPoint values.

        The shift values should positive or negative integer:

        - if the value is positive, it is a left-shift,
        - if the value is negative, it is a right-shift.

        The resulting FixedPoint has the same value bits and fractional bits as
        the source FixedPoint, which means that clipping is applied on
        left-shift and flooring is applied on right-shift.

        Args:
            s (:obj:`tensorflow.Tensor`): the shift values for each pixel.

        Returns:
            :obj:`FixedPoint`: the result as a FixedPoint
        """
        values = floor_through(self.values * pow2(s))
        return FixedPoint(values, self.frac_bits, self.value_bits)

    def __rshift__(self, shift):
        """Right shift the FixedPoint values

        This operation has no direct equivalent in float arithmetics: it corresponds to a division
        of the corresponding float by a power-of-two, then a flooring to the quantization interval.

        Args:
            shift (:obj:`tensorflow.Tensor`): the power of 2 to divide by

        Returns:
            :obj:`FixedPoint`: the result as a FixedPoint
        """
        shift = tf.cast(shift, tf.float32)
        assert_less_equal(0, shift, "Shift must be all positive")
        assert_equal(tf.rank(shift) <= tf.rank(self.frac_bits), True,
                     "The shift's rank must be less than or equal to the rank of frac_bits. "
                     f"Received {tf.rank(shift)} > {tf.rank(self.frac_bits)}."
                     "That means it is not possible to fold the shift into the FixedPoint. "
                     "Please use FixedPoint.shift instead of '>>'.")

        # The shift can be folded into the fractional bits,
        s_frac_bits = self.frac_bits + shift
        # keeping the same values
        s_values = self.values
        # Return a new FixedPoint with updated fractional bits,
        # which is equivalent in hardward without performing any operation
        return FixedPoint(s_values, s_frac_bits, self.value_bits)

    def __lshift__(self, shift):
        """Left shift the FixedPoint values

        This operation has no direct equivalent in float arithmetics: it corresponds to a
        multiplication of the corresponding float by a power-of-two, then a flooring to the
        quantization interval.

        Args:
            shift (:obj:`tensorflow.Tensor`): the power of 2 to multiply by

        Returns:
            :obj:`FixedPoint`: the result as a FixedPoint
        """
        assert_less_equal(0, shift, "Shift must be all positive")
        # Simply apply the shift on the values
        s_values = FixedPoint._lshift(self.values, shift)
        # Return a new FixedPoint with updated values
        return FixedPoint(s_values, self.frac_bits, self.value_bits)

    def _align_values(self, other):
        # The sub fractional bits are the max of both terms
        frac_bits = tf.math.maximum(self.frac_bits, other.frac_bits)
        self_values = FixedPoint._lshift(
            self.values, (frac_bits - self.frac_bits))
        other_values = FixedPoint._lshift(
            other.values, (frac_bits - other.frac_bits))
        return frac_bits, self_values, other_values

    def __add__(self, other):
        if isinstance(other, int):
            # Convert integer into a 32-bit fixed-point with no fractional bits,
            # aligned with the current FixedPoint
            return self + FixedPoint.quantize(other, 32, self.frac_bits)
        elif isinstance(other, FixedPoint):
            # Check that self and other are per-tensor
            self.assert_per_tensor()
            other.assert_per_tensor()
            # Check that self and other are aligned
            assert_equal(self.frac_bits, other.frac_bits)
            # Return a new FixedPoint
            return FixedPoint(self.values + other.values, self.frac_bits, self.value_bits)
        raise TypeError(
            f"Unsupported operand type(s) for +: 'FixedPoint' and '{type(other)}'")

    def __sub__(self, other):
        if isinstance(other, int):
            # Convert integer into a 32-bit fixed-point with no fractional bits,
            # aligned with the current FixedPoint
            return self - FixedPoint.quantize(other, 32, self.frac_bits)
        elif isinstance(other, FixedPoint):
            # Check that self and other are per-tensor
            self.assert_per_tensor()
            other.assert_per_tensor()
            # Check that self and other are aligned
            assert_equal(self.frac_bits, other.frac_bits)
            # Return a new FixedPoint
            return FixedPoint(self.values - other.values, self.frac_bits, self.value_bits)
        raise TypeError(
            f"Unsupported operand type(s) for -: 'FixedPoint' and '{type(other)}'")

    def __mul__(self, other):
        if isinstance(other, int):
            return self * FixedPoint(other)
        elif isinstance(other, FixedPoint):
            # The product between fixed-point is straight-forward
            p_values = self.values * other.values
            # Return a new FixedPoint whose frac bits is the sum of both terms
            return FixedPoint(p_values, self.frac_bits + other.frac_bits, self.value_bits)
        raise TypeError(
            f"Unsupported operand type(s) for *: 'FixedPoint' and '{type(other)}'")

    def __truediv__(self, other):
        if isinstance(other, int):
            return self / FixedPoint(other)
        elif isinstance(other, FixedPoint):
            assert_none_equal(other.values, 0.0, "Cannot divide by 0.")
            # The division between fixed-point is straight-forward
            d_values = floor_through(self.values / other.values)
            # Return a new FixedPoint whose frac bits from other is subtracted.
            return FixedPoint(d_values, self.frac_bits - other.frac_bits, self.value_bits)
        raise TypeError(
            f"Unsupported operand type(s) for /: 'FixedPoint' and '{type(other)}'")

    def __pow__(self, power):
        if isinstance(power, int):
            if power == 0:
                return FixedPoint(tf.ones_like(self.values), 0, self.value_bits)
            elif power == 1:
                return FixedPoint(self.values, self.frac_bits, self.value_bits)
            elif power > 1:
                return self * self ** (power - 1)
            else:
                raise NotImplementedError(
                    "Negative powers are not implemented yet")
        raise TypeError(
            f"Unsupported operand type(s) for **: 'FixedPoint' and '{type(power)}'")

    def __gt__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for >: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values > o_values

    def __ge__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for >=: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values >= o_values

    def __eq__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for ==: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values == o_values

    def __ne__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"unsupported operand type(s) for !=: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values != o_values

    def __lt__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for <: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values < o_values

    def __le__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for <=: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values <= o_values

    def abs(self):
        """Returns the absolute value of the FixedPoint

        Returns:
            :obj:`FixedPoint`: the absolute value.
        """
        return FixedPoint(tf.math.abs(self.values), self.frac_bits, self.value_bits)

    def floor(self):
        """Floors the FixedPoint

        Returns:
            tuple(:obj:`FixedPoint`, :obj:`tf.Tensor`): a new FixedPoint without
                fractional bits and the shift that was applied.
        """
        # Divide values to remove fractional bits
        values = FixedPoint._rshift(self.values, self.frac_bits)
        return FixedPoint(values, 0, self.value_bits), self.frac_bits



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import keras
import tensorflow as tf
import numpy as np
from typing import Callable
import warnings

_GLOBAL_CUSTOM_OBJECTS = {}


class StaticHashTable(keras.layers.Layer):
    """Reference object that extends the :obj:`tf.lookup.StaticHashTable` declaration,
    in order to store the input data in independent variables and **build the static table at
    time of building the layer**, looking to manipulate the layer as an general object.

    Args:
        keys (:obj:`np.ndarray`): the keys tensor in the static table.
        values (:obj:`np.ndarray`): the values tensor in the static table.
        default_value (:obj:`np.ndarray`): the default value.
        name (str, optional): the name of the operation. Defaults to None.
    """

    def __init__(self, keys, values, default_value, name=None, **kwargs):
        super().__init__(name=name, **kwargs)
        if not (isinstance(keys, np.ndarray) and isinstance(values, np.ndarray)):
            raise ValueError("StaticHashTable does not support Tensor. Please use numpy values.")
        self._keys, self._values = keys, values
        self._kwargs = {'default_value': default_value}

    def build(self, _):
        self.hash_table = tf.lookup.StaticHashTable(tf.lookup.KeyValueTensorInitializer(
            self._keys, self._values, self._keys.dtype, self._values.dtype), **self._kwargs)

    def call(self, x):
        return self.hash_table.lookup(tf.cast(x, self.hash_table.key_dtype))


class RegisterCustomHashTable:
    """Register values generated by a function in a global variable, as a static hash table.

    Given a generator function, :obj:`RegisterCustomHashTable` will store the data generated by it
    as a global variable, identified with a table name.

    The following code is an example of the common uses:

    .. code-block:: python
        with RegisterCustomHashTable(table_name='my_table', gen_func=my_func) as register:
            table = register.get_table()

    Args:
        table_name (str): the name of the hash table to get.
        gen_func (Callable): the generator function. Expected to return a tuple of three elements:
            - keys (:obj:`np.ndarray`): the input tensor.
            - values (:obj:`np.ndarray`): the output tensor.
            - default_value (:obj:`np.ndarray`): the default value.
        drop_table (bool, optional): if True, the hash table will be dropped after the context is
            exited. Defaults to False.
        *args, **kwargs: additional arguments used in the call of generator function.

    Note:
        The generator function only will be called if the hash table is not registered.

    Raises:
        ValueError: if the generator function does not return a tuple of three elements.
    """

    def __init__(self, table_name: str, gen_func: Callable, *args, drop_table=False, **kwargs):
        self.table_name = table_name
        self.gen_func = gen_func
        self.drop_table = drop_table
        self._args = args
        self._kwargs = kwargs

    def get_table(self):
        """Return the hash table given a table name setting in context.

        Returns:
            :obj:`tf.lookup.StaticHashTable`: the static hash table.
        """
        return self.get_static_hash_table(self.table_name)

    @staticmethod
    def get_static_hash_table(table_name: str):
        """Return the hash table registered previously.

        Args:
            table_name (str): the name of the hash table.

        Returns:
            :obj:`tf.lookup.StaticHashTable`: the static hash table.
        """
        # Retrieves hash static table name
        if table_name not in _GLOBAL_CUSTOM_OBJECTS:
            raise ValueError(f'Hash table {table_name} not found')
        return _GLOBAL_CUSTOM_OBJECTS[table_name]

    @staticmethod
    def drop_hash_table(table_name: str):
        """Drop a hash table given its name.

        Args:
            table_name (str): the name of the hash table.
        """
        if table_name not in _GLOBAL_CUSTOM_OBJECTS:
            raise ValueError(f'Hash table {table_name} not found')
        _GLOBAL_CUSTOM_OBJECTS.pop(table_name)

    def _register_static_hash_table(self):
        """ Register a static hash table, where the data is generated by :attr:`gen_func`.
        """
        if self.table_name in _GLOBAL_CUSTOM_OBJECTS:
            warnings.warn(f'Hash table {self.table_name} already exists. Overwriting...')

        # Register hash table
        keys, values, default_value = self.gen_func(*self._args, **self._kwargs)
        _GLOBAL_CUSTOM_OBJECTS[self.table_name] = StaticHashTable(
            keys, values, name=self.table_name, default_value=default_value)

    def _check_registered_hash_table(self):
        """Check if a hash table is registered given its name.

        Returns:
            bool: True if the hash table is registered, False otherwise.
        """
        return self.table_name in _GLOBAL_CUSTOM_OBJECTS

    def __enter__(self):
        """ Builtin method to enter of custom context scope ('with' statement),
        registering a hash table if it is not in the context.
        """
        if not self._check_registered_hash_table():
            self._register_static_hash_table()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Builtin method to exit of custom context scope ('with' statement) and
        drop the hash table if :attr:`drop_table` is True.
        """
        if self.drop_table and self._check_registered_hash_table():
            self.drop_hash_table(self.table_name)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Common utility methods used in quantization models.
"""

from copy import deepcopy
from pathlib import Path
import numpy as np
import tensorflow_addons as tfa
from keras.models import clone_model, load_model as kload_model, Model
import warnings

from quantizeml.layers.quantizers import Quantizer, Dequantizer

from ..layers import Calibrable

__all__ = ['load_model', 'deep_clone_model', 'set_calibrate', 'insert_layer', 'load_weights',
           'save_weights']


def load_model(model_path, custom_layers={}, compile_model=True):
    """Loads a model with Vision Transformer custom layers.

    Args:
        model_path (str): path of the model to load
        custom_layers (dict, optional): custom layers to add to the model. Defaults to {}.
        compile_model (bool, optional): whether to compile the model. Defaults to True.

    Returns:
        :class:`keras.models.Model`: the loaded model
    """
    # Given that all layers were imported in __init__.py, they are registered.
    # Therefore, we just need append GroupNormalization in the custom layers.
    custom_layers.update({'GroupNormalization': tfa.layers.GroupNormalization})
    return kload_model(model_path, custom_objects=custom_layers, compile=compile_model)


def deep_clone_model(model, *args, **kwargs):
    """Clone a model, assign variable to variable. Useful when a clone function is used,
    and new layers have not the same number of parameters as the original layer.

    Args:
        model (:class:`keras.models.Model`): model to be cloned
        args, kwargs (optional): arguments pass to :func:`keras.models.clone_model` function

    Returns:
        :class:`keras.models.Model`: the cloned model
    """
    new_model = clone_model(model, *args, **kwargs)
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(new_model, variables_dict, False)
    return new_model


def set_calibrate(model, value):
    """Enable or disable calibration for all layers in a model.

    Args:
        model (:class:`keras.models.Model`): model to be calibrated
        value (bool): True to enable calibration, False to disable it
    """
    for layer in model.layers:
        if isinstance(layer, Calibrable):
            layer.calibration = value


def insert_layer(model, target_layer_name, new_layer):
    """ Inserts the given layer in the model before or after the layer with the name
    target_layer_name depending on the type of new_layer.

    Note that new_layer type is restricted to (Quantizer, Dequantizer) and a Quantizer will be added
    before new_layer while a Dequantizer will be added after

    Args:
        model (keras.Model): the model to update
        target_layer_name (str): name of the layer after which to insert a layer
        new_layer (keras.Layer): layer to insert

    Raises:
        ValueError: when target_layer_name is not found in model or new_layer is not in (Quantizer,
            Dequantizer)

    Returns:
        keras.Model: the new model
    """
    # Check that the model has a layer with then given target_layer_name
    if not any(ly.name == target_layer_name for ly in model.layers):
        raise ValueError(f'{target_layer_name} not found in model.')

    # Check added layer type
    if not isinstance(new_layer, (Quantizer, Dequantizer)):
        raise ValueError(f'Inserted layer must be of type Quantizer or Dequantizer, \
                        received {type(new_layer)}.')

    # Create a clone model to prevent messing with the original model
    clone_model = deep_clone_model(model)

    # Auxiliary dictionary to describe the network graph
    network_dict = {'input_layers_of': {}, 'new_output_tensor_of': {}}

    # Set the input layers of each layer
    for layer in clone_model.layers:
        for node in layer._outbound_nodes:
            layer_name = node.outbound_layer.name
            if layer_name not in network_dict['input_layers_of']:
                network_dict['input_layers_of'].update({layer_name: [layer.name]})
            else:
                network_dict['input_layers_of'][layer_name].append(layer.name)

    # Set the output tensor of the input layer
    network_dict['new_output_tensor_of'].update({clone_model.layers[0].name: clone_model.input})

    # Iterate over all layers after the input
    model_outputs = []
    insert_before = isinstance(new_layer, Quantizer)
    for layer in clone_model.layers[1:]:
        # Determine input tensors
        layer_input = []
        for layer_aux in network_dict['input_layers_of'][layer.name]:
            input = network_dict['new_output_tensor_of'][layer_aux]
            if isinstance(input, tuple):
                network_dict['new_output_tensor_of'][layer_aux] = input[1:]
                input = input[0]
            layer_input.append(input)
        if len(layer_input) == 1:
            layer_input = layer_input[0]

        # Insert layer if name matches the given layer name
        if layer.name == target_layer_name:
            if insert_before:
                # Apply one Quantizer by input then current layer
                if isinstance(layer_input, list):
                    inputs = []
                    for i in range(len(layer_input)):
                        copied_layer = deepcopy(new_layer)
                        copied_layer._name += "_" + str(i)
                        y = copied_layer(layer_input[i])
                        inputs.append(y)
                    x = inputs
                else:
                    x = new_layer(layer_input)
                x = layer(x)
            else:
                # Apply current layer then new layer
                x = layer(layer_input)
                if isinstance(x, tuple):
                    x = tuple(new_layer(i) for i in x)
                else:
                    x = new_layer(x)
        else:
            x = layer(layer_input)

        # Set new output tensor (the original one, or the one of the inserted layer)
        network_dict['new_output_tensor_of'].update({layer.name: x})

        # Save tensor in output list if it is output in initial model
        if layer.name in clone_model.output_names:
            model_outputs.append(x)

    new_model = Model(inputs=clone_model.inputs, outputs=model_outputs)
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(new_model, variables_dict, False)
    return new_model


def apply_weights_to_model(model, weights, verbose=True):
    """Loads weights from a dictionary and apply it to a model.

    Go through the dictionary of weights, find the corresponding variable in the
    model and partially load its weights.

    Args:
        model (keras.Model): the model to update
        weights (dict): the dictionary of weights
        verbose (bool, optional): if True, throw warning messages if a dict item is not found in the
            model. Defaults to True.
    """
    if len(weights) == 0:
        warnings.warn("There is no weight to apply to the model.")
        return

    # Go through the dictionary of weights with each item
    for key, value in weights.items():
        value_applied = False
        for dest_var in model.variables:
            if key == dest_var.name:
                # Apply the current item value
                dest_var.assign(value)
                value_applied = True
                break
        if not value_applied and verbose:
            warnings.warn(f"Variable '{key}' not found in the model.")


def load_weights(model, weights_path):
    """Loads weights from a npz file and apply it to a model.

    Go through the dictionary of weights of the npz file, find the corresponding variable in the
    model and partially load its weights.

    Args:
        model (keras.Model): the model to update
        weights_path (str): the path of the npz file to load
    """
    # Check the npz file validity
    path = Path(weights_path)
    if not path.is_file():
        raise ValueError(f"File `{weights_path}` not found.")

    # Open the npz file
    weights_dict = np.load(weights_path)

    # Apply the weights to the model
    apply_weights_to_model(model, weights_dict)


def save_weights(model, weights_path):
    """Save model weights on an npz file.

    Takes a model and save the weights of all its layers into an npz file.

    Args:
        model (keras.Model): the model to save its weights
        weights_path (str): the path of the npz file to save
    """
    weights_dict = {}
    for var in model.variables:
        weights_dict[var.name] = var

    np.savez(weights_path, **weights_dict)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import keras
import tensorflow as tf

from ..tensors import FixedPoint


__all__ = ["QDropout"]


@tf.keras.utils.register_keras_serializable()
class QDropout(keras.layers.Dropout):
    """ A dropout layer that operates on quantized inputs and weights.

    It is only implemented as a passthrough.
    """

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QDropout only accepts FixedPoint inputs.\
                             Receives {type(inputs)} inputs.")

        # QuantizedDropout act as a pass through.
        return inputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
import numpy as np
from typing import Union
from .qtensor import pow2
from .fixed_point import FixedPoint
from .lookup import RegisterCustomHashTable

QTENSOR_T = Union[FixedPoint, tf.Tensor]


def _get_static_hash_table(range_bitwidth: int, frac_bits: int):
    """Given an input range bitwitdh, register a static hash table which contains
    the reciprocal values between (-2**range_bitwidth) and (2**range_bitwidth - 1).

    Args:
        range_bitwidth (int): the bitwidth of the range of the table.
        frac_bits (int): the resolution of fractional bits.

    Returns:
        :obj:`tf.lookup.StaticHashTable`: the static hash table.
    """
    def _serialize_hash_table():
        # Given that StaticHashTable does not support Tensor, all the values are computed in
        # numpy directly.
        keys_tensor = np.array(range(1, 2**range_bitwidth), dtype=np.int32)
        vals_tensor = np.floor(1 / keys_tensor * 2.0 ** frac_bits).astype(np.float32)
        return keys_tensor, vals_tensor, 0

    # Register the table if it does not exist as a global variable
    table_name = f'reciprocal_lut_{frac_bits}_{range_bitwidth}bits'
    with RegisterCustomHashTable(table_name, _serialize_hash_table) as register:
        table = register.get_table()
    return table


def reciprocal_lut(
        x: QTENSOR_T, frac_bits: int = None, out_value_bits: int = 32, name: str = None):
    """Compute the reciprocal of a FixedPoint value, using a lookup table.

    Args:
        x (:obj:`FixedPoint` or :obj:`tf.Tensor`): the value to compute its reciprocal.
        frac_bits (int, optional): the resolution of fractional bits. Defaults to None.
        out_value_bits (int, optional): the number of bits of the output. Defaults to 32.
        name (str, optional): the name of the operation. Defaults to None.

    Returns:
        :obj:`FixedPoint` or :obj:`tf.Tensor`: the reciprocal of x.
    """
    # Retrieves the reciprocal lookup table only if x is a FixedPoint
    if not isinstance(x, FixedPoint):
        return tf.math.reciprocal(x)

    # Retrieves the hash static table of x.value_bits
    frac_bits = frac_bits or x.frac_bits
    hash_table = _get_static_hash_table(x.value_bits, frac_bits)

    # Given that the input is a FixedPoint, it is guarantee the equivalence below:
    # float(1/x) = 1 / float(x) =  1/x.values * 2^x.frac_bits
    # Then, we can approximate the reciprocal of x to:
    # float(1/x) ~= reciprocal_lut(x.values) * 2^x.frac_bits
    # in two steps:
    # 1. Compute the reciprocal of x.values of absolute value
    x_inv = FixedPoint(hash_table(x.abs().values), frac_bits, out_value_bits)
    # 2. Scale by 2**x.frac_bits
    return x_inv * FixedPoint(pow2(x.frac_bits)) * x.sign



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
QuantizedSeparableConv2D layer definition.
"""

import tensorflow as tf
from keras.layers import SeparableConv2D
from keras.utils import conv_utils
from keras import backend

from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH, QFloat
from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer


__all__ = ["QuantizedSeparableConv2D", "SeparableConv2DTranspose",
           "QuantizedSeparableConv2DTranspose"]


@tf.keras.utils.register_keras_serializable()
class QuantizedSeparableConv2D(Calibrable, SeparableConv2D):
    """ A separable convolutional layer that operates on quantized inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        if 'dilation_rate' in kwargs:
            if kwargs['dilation_rate'] not in [1, [1, 1], (1, 1)]:
                raise ValueError("Keyword argument 'dilation_rate' is not supported in \
                                 QuantizedSeparableConv2D.")
        if 'depth_multiplier' in kwargs:
            if kwargs['depth_multiplier'] != 1:
                raise ValueError("Keyword argument 'depth_multiplier' is not supported in \
                                 QuantizedSeparableConv2D.")

        super().__init__(*args, **kwargs)
        self.quant_config = quant_config

        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)

        # Separable layer has two weights quantizers to handle different max values
        self.dw_weight_quantizer = deserialize_quantizer(
            self.quant_config, "dw_weight_quantizer", WeightQuantizer, True)
        if self.dw_weight_quantizer.get_config()['axis'] != "per-tensor":
            raise ValueError("Separable depthwise weights must be quantized per tensor")
        self.pw_weight_quantizer = deserialize_quantizer(
            self.quant_config, "pw_weight_quantizer", WeightQuantizer, True)

        if self.use_bias:
            self.bias_quantizer = deserialize_quantizer(
                self.quant_config, "bias_quantizer", WeightQuantizer, True)

        self.buffer_bitwidth = self.quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."

        if self.use_bias:
            # Add objects that will store the shift values.
            self.bias_shift = CalibrableVariable()
            self.output_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint or tf.Tensor
        if not isinstance(inputs, (FixedPoint, tf.Tensor)):
            raise TypeError(f"QuantizedSeparableConv2D only accepts FixedPoint\
                               or tf.Tensor inputs. Receives {type(inputs)} inputs.")

        if isinstance(inputs, tf.Tensor):
            # Assume the inputs are integer stored as float, which is the only tf.Tensor
            # inputs that are allowed
            inputs = FixedPoint(inputs, 0, 8)

        # Although the dephwise operation does not require it, we only accept inputs quantized
        # per-tensor to avoid increasing too much the fractional bits of the depthwise outputs.
        inputs.assert_per_tensor()

        # Quantize the weights
        depthwise_kernel = self.dw_weight_quantizer(self.depthwise_kernel, training)
        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel, training)

        inputs = inputs.promote(self.buffer_bitwidth)
        dw_outputs_q = backend.depthwise_conv2d(
            inputs,
            depthwise_kernel,
            strides=self.strides,
            padding=self.padding,
            dilation_rate=self.dilation_rate,
            data_format=self.data_format)

        outputs = tf.nn.convolution(
            dw_outputs_q,
            pointwise_kernel,
            strides=[1, 1, 1, 1],
            padding='VALID',
            data_format=conv_utils.convert_data_format(self.data_format, ndim=4))

        if self.use_bias:
            bias = self.bias_quantizer(self.bias, training).promote(self.buffer_bitwidth)

            # Align intermediate outputs and biases before adding them
            outputs, shift = outputs.align(bias)
            self.output_shift(shift)
            bias, shift = bias.align(outputs)
            self.bias_shift(shift)
            outputs = tf.add(outputs, bias)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config


@tf.keras.utils.register_keras_serializable()
class SeparableConv2DTranspose(SeparableConv2D):
    """ A transposed separable convolutional layer.

    It performs a transposed depthwise convolution on inputs followed by a standard pointwise
    operation.
    """

    def __init__(self, *args, **kwargs):
        if 'dilation_rate' in kwargs:
            if kwargs['dilation_rate'] not in [1, [1, 1], (1, 1)]:
                raise ValueError("Keyword argument 'dilation_rate' is not supported in "
                                 "SeparableConv2DTranspose.")
        if 'depth_multiplier' in kwargs:
            if kwargs['depth_multiplier'] != 1:
                raise ValueError("Keyword argument 'depth_multiplier' is not supported in "
                                 "SeparableConv2DTranspose.")
        # Limit supported stride to 2. Standard separable should be used for stride 1 and greater
        # strides are not supported.
        if 'strides' in kwargs:
            if kwargs['strides'] not in [2, [2, 2], (2, 2)]:
                raise ValueError(f"Only supported stride is 2. Received {kwargs['strides']}.")
        # Also limit padding to 'same'
        if 'padding' in kwargs:
            if kwargs['padding'] != 'same':
                raise ValueError(f"Only supported padding is 'same'. Received {kwargs['padding']}.")
        super().__init__(*args, **kwargs)

    def call(self, inputs):
        # Infer the dynamic output shape
        inputs_shape = tf.shape(inputs)
        out_height = conv_utils.deconv_output_length(inputs_shape[1],
                                                     self.kernel_size[0],
                                                     padding=self.padding,
                                                     stride=self.strides[0],
                                                     dilation=self.dilation_rate[0])
        out_width = conv_utils.deconv_output_length(inputs_shape[2],
                                                    self.kernel_size[1],
                                                    padding=self.padding,
                                                    stride=self.strides[1],
                                                    dilation=self.dilation_rate[1])
        output_shape = tf.stack((inputs_shape[0], out_height, out_width, 1))

        # Inputs and kernels must be transposed to have their channel dimension first because the
        # tf.vectorized_map call that follows will unpack them on dimension 0. The channel dimension
        # is virtually restored using expand_dims so that elements have the appropriate shape for
        # the conv2d_transpose call (with a channel dimension of 1 which is expected in the
        # depthwise process).
        inputs_channel_first = tf.transpose(inputs, (3, 0, 1, 2))
        inputs_channel_first = tf.expand_dims(inputs_channel_first, -1)
        kernel_channel_first = tf.transpose(self.depthwise_kernel, (2, 0, 1, 3))
        kernel_channel_first = tf.expand_dims(kernel_channel_first, -2)

        dw_outputs = tf.vectorized_map(
            lambda x: backend.conv2d_transpose(x[0],
                                               x[1],
                                               output_shape=output_shape,
                                               strides=self.strides,
                                               padding=self.padding),
            (inputs_channel_first, kernel_channel_first))
        dw_outputs = tf.transpose(tf.squeeze(dw_outputs, axis=-1), (1, 2, 3, 0))

        # Pointwise operation
        outputs = tf.nn.convolution(
            dw_outputs,
            self.pointwise_kernel,
            strides=[1, 1, 1, 1],
            padding='VALID')

        if self.use_bias:
            outputs = tf.add(outputs, self.bias)
        return outputs


@tf.keras.utils.register_keras_serializable()
class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
    """ A transposed separable convolutional layer that operates on quantized inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config

        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)

        # Transposed Separable layer has two weights quantizers to handle different max values
        self.dw_weight_quantizer = deserialize_quantizer(
            self.quant_config, "dw_weight_quantizer", WeightQuantizer, True)
        self.pw_weight_quantizer = deserialize_quantizer(
            self.quant_config, "pw_weight_quantizer", WeightQuantizer, True)

        # Depthwise quantizer must be per-tensor
        if "axis" in self.quant_config["dw_weight_quantizer"]:
            if self.quant_config["dw_weight_quantizer"]["axis"] != "per-tensor":
                raise ValueError("Only supporting 'per-tensor' depthwise quantizer. Received "
                                 f"{self.quant_config['dw_weight_quantizer']['axis']}.")

        if self.use_bias:
            self.bias_quantizer = deserialize_quantizer(
                self.quant_config, "bias_quantizer", WeightQuantizer)

        self.buffer_bitwidth = self.quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."
        self.intermediate_quantizer = deserialize_quantizer(
            self.quant_config, "intermediate_quantizer", OutputQuantizer, False)

        if self.intermediate_quantizer:
            intermediate_quantizer_axis = self.intermediate_quantizer.get_config()['axis']
            if intermediate_quantizer_axis != "per-tensor":
                raise ValueError("Only supporting 'per-tensor' intermediate quantizer. Received "
                                 f"{intermediate_quantizer_axis}.")

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        self.pw_shift = CalibrableVariable()
        self.bias_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint or tf.Tensor
        if not isinstance(inputs, (FixedPoint, tf.Tensor)):
            raise TypeError("QuantizedSeparableConv2DTranspose only accepts FixedPoint or "
                            f"tf.Tensor inputs. Receives {type(inputs)} inputs.")

        if isinstance(inputs, tf.Tensor):
            # Assume the inputs are integer stored as float, which is the only tf.Tensor
            # inputs that are allowed
            inputs = FixedPoint.quantize(inputs, 0, 8)

        # Promote inputs to avoid saturation, then align them
        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
        self.input_shift(shift)

        # Infer the dynamic output shape
        inputs_shape = tf.shape(inputs)
        out_height = conv_utils.deconv_output_length(inputs_shape[1],
                                                     self.kernel_size[0],
                                                     padding=self.padding,
                                                     stride=self.strides[0],
                                                     dilation=self.dilation_rate[0])
        out_width = conv_utils.deconv_output_length(inputs_shape[2],
                                                    self.kernel_size[1],
                                                    padding=self.padding,
                                                    stride=self.strides[1],
                                                    dilation=self.dilation_rate[1])
        output_shape = tf.stack((inputs_shape[0], out_height, out_width, 1))

        # Inputs and kernels must be transposed to have their channel dimension first because the
        # tf.vectorized_map call that follows will unpack them on dimension 0. The channel dimension
        # is virtually restored using expand_dims so that elements have the appropriate shape for
        # the conv2d_transpose call (with a channel dimension of 1 which is expected in the
        # depthwise process).
        inputs_channel_first = tf.transpose(inputs, (3, 0, 1, 2))
        inputs_channel_first = tf.expand_dims(inputs_channel_first, -1)
        kernel_channel_first = tf.transpose(self.depthwise_kernel, (2, 0, 1, 3))
        kernel_channel_first = tf.expand_dims(kernel_channel_first, -2)

        # Quantize the depthwise kernels
        depthwise_kernel = self.dw_weight_quantizer(kernel_channel_first, training)

        # Perform the depthwise operation on values using conv2d_transpose on each channel
        dw_values = tf.vectorized_map(
            lambda x: backend.conv2d_transpose(x[0],
                                               x[1],
                                               output_shape=output_shape,
                                               strides=self.strides,
                                               padding=self.padding),
            (inputs_channel_first.values, depthwise_kernel.values))
        dw_values = tf.transpose(tf.squeeze(dw_values, axis=-1), (1, 2, 3, 0))

        if isinstance(depthwise_kernel, QFloat):
            # Multiply by the scale
            dw_values *= depthwise_kernel.scales.values

        # Build a new FixedPoint
        if isinstance(depthwise_kernel, FixedPoint):
            filters_frac_bits = depthwise_kernel.frac_bits
        else:
            filters_frac_bits = depthwise_kernel.scales.frac_bits
        dw_outputs = FixedPoint(dw_values, inputs.frac_bits + filters_frac_bits, inputs.value_bits)

        if self.intermediate_quantizer is not None:
            dw_outputs_q = self.intermediate_quantizer(dw_outputs, training)
            dw_outputs_q = dw_outputs_q.promote(self.buffer_bitwidth)
        else:
            dw_outputs_q = dw_outputs

        # Quantize the pointwise kernel
        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel, training)

        # Pointwise operation
        outputs = tf.nn.convolution(
            dw_outputs_q,
            pointwise_kernel,
            strides=[1, 1, 1, 1],
            padding='VALID')

        if self.use_bias:
            # Quantize biases
            bias = self.bias_quantizer(self.bias, training)
            # Align intermediate outputs and biases before adding them
            outputs, shift = outputs.align(bias)
            self.pw_shift(shift)
            bias, shift = bias.promote(self.buffer_bitwidth).align(outputs)
            self.bias_shift(shift)
            outputs = tf.add(outputs, bias)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from contextlib import contextmanager
import tensorflow as tf
import keras.backend as K
from keras.layers import Layer

from ..tensors import QTensor, QFloat, FixedPoint, ceil_log2
from .calibrable import Calibrable, CalibrableVariable


__all__ = ["deserialize_quantizer", "Quantizer", "Dequantizer", "WeightQuantizer",
           "OutputQuantizer"]


def deserialize_quantizer(quant_config, name, quantizer_class, mandatory=True):
    """Wrapper function of tf.keras.utils.deserialize_keras_object.

    It allows to select the right config from the config file dict,
    and raises an error if no config found or set to None.
    If one is found, it returns the corresponding deserialized Quantizer.

    Args:
        quant_config (dict): quantization config dictionnary.
        name (str): keras object name to deserialize.
        quantizer_class (object): the Quantizer subclass to instantiate.
        mandatory (bool, optional): flag to specify if the object to
            deserialize is mandatory. If yes raises an Error otherwise
            returns None. Defaults to True.

    Returns:
        :obj:`Quantizer`: the targeted deserialized Quantizer
        object.
    """
    object_dict = quant_config.get(name)
    if object_dict is None:
        if mandatory:
            raise KeyError(f"'{name}' should be specified.")
        return None

    # Check that object_dict has the right keys
    list_available_keys = {
        "trainable": bool,
        "bitwidth": int,
        "signed": bool,
        "axis": str,
        "momentum": float,
        "dtype": str,
        "scale_bits": int
    }
    for key in object_dict:
        if key == "scale_bits" and quantizer_class == OutputQuantizer:
            raise ValueError(f"'{name}' produces FixedPoint output. "
                             "It doesn't support 'scale_bits' key.")
        if key not in list_available_keys:
            raise KeyError(f"'{key}' is not a valid key for '{name}'.")
        if not isinstance(object_dict[key], list_available_keys[key]):
            raise ValueError(f"'{key}' should be of type {list_available_keys[key]}. "
                             f"Received {type(object_dict[key])}.")

    # Deserialize the object
    deserialize_dict = object_dict.copy()
    deserialize_dict["name"] = name
    return quantizer_class.from_config(deserialize_dict)


@contextmanager
def disable_partitioner(layer):
    partitioner = None
    try:  # Disable variable partitioning when creating the moving tensors
        if hasattr(layer, "_scope") and layer._scope:
            partitioner = layer._scope.partitioner
            layer._scope.set_partitioner(None)
        yield layer
    finally:  # Restore partitioner
        if partitioner:
            layer._scope.set_partitioner(partitioner)


class Quantizer(Layer):
    """The base class for all quantizers.

    The bitwidth defines the number of quantization levels on which the
    values will be quantized.
    For a quantizer that accepts unsigned values, the maximum quantization
    level is 2 ^ bitwidth - 1.
    For a quantizer that accepts signed values, we lose one bit of precision to
    store the sign.
    When the quantizer is signed, the quantization interval is asymmetric around
    zero (i.e range: [- 2 ^ (bitwidth - 1), 2 ^ (bitwidth - 1) - 1]).
    The quantization is actually performed on absolute values, between 0 and
    max_value, where:

    - max_value is either a scalar (per-tensor quantization), or a vector
      (per-axis quantization),
    - max_value is a static value on inference, and an adaptative value on training.

    Args:
        bitwidth (int): the quantization bitwidth.
        signed (bool, optional): whether the quantizer expects signed values or unsigned.
            Defaults to True.
        axis (str, optional): reduce across all tensor values ('per-tensor') or keep the
            last axis ('per-axis'). Defaults to 'per-tensor'.

    Note:
        To get more information about the moving average implementation, see the
        `BatchNormalizationLayer <https://bit.ly/3KcEaUh>`_ class.
    """

    def __init__(self, bitwidth, signed=True, axis="per-tensor", **kwargs):
        assert bitwidth > 1
        self.bitwidth = bitwidth
        self.signed = signed
        self.value_bits = bitwidth - 1 if signed else bitwidth
        self._axis = axis
        if not (isinstance(axis, str) and axis in ["per-tensor", "per-axis"]):
            raise ValueError(
                f"Only support reduction 'per-tensor' or 'per-axis'. Given {axis}.")
        super().__init__(**kwargs)

    def build(self, input_shape):
        """Build the layer.

        Args:
            input_shape (list): the shape of input tensor.
        """
        # Convert axis to a list of int
        if self._axis == "per-axis" and len(input_shape) > 1:
            self.axis = list(range(len(input_shape) - 1))
        else:
            self.axis = None

    def frac_bits(self, max_value):
        # Clamp the max_value to the next power-of-two
        int_bits = tf.cast(ceil_log2(max_value), tf.int32)
        int_bits = tf.clip_by_value(int_bits, 0, self.value_bits)
        # Evaluate fractional bits
        return self.value_bits - int_bits

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"bitwidth": self.bitwidth})
        config.update({"signed": self.signed})
        config.update({"axis": self._axis})
        return config


@tf.keras.utils.register_keras_serializable()
class WeightQuantizer(Quantizer):
    """A trivial uniform quantizer that has only one scale for everything.

    Scale is dynamic (depends on inputs), and it can be read via the scale property.

    Args:
        bitwidth (int): the quantization bitwidth.
        scale_bits (int, optional): the number of bits for the scaling. Defaults to None.
    """

    def __init__(self, bitwidth, scale_bits=None, **kwargs):
        if not isinstance(bitwidth, int) or bitwidth < 2:
            raise ValueError(
                f"Bitwidth should be an int >= 2, currently {bitwidth}")
        self.scale_bits = scale_bits
        super().__init__(bitwidth, **kwargs)

    def call(self, inputs, training=None):
        """Update the max_value from the new inputs, and quantize them in three steps:

            1. Check if the inputs are float values,
            2. Compute the max_value,
            3. Quantize the inputs.

        Args:
            inputs (:obj:`tensorflow.Tensor` or :obj:`QTensor`): the inputs tensor.
            training (bool, optional): the training mode. Defaults to None.

        Returns:
            :obj:`tensorflow.Tensor`: the quantized tensor.
        """
        if isinstance(inputs, QTensor):
            raise ValueError(
                f"{type(inputs)} input is not supported. WeightQuantizer only accepts float"
                " inputs.")

        # Compute the max_value from the inputs
        max_value = tf.math.reduce_max(tf.math.abs(inputs), self.axis)
        # Return a QFloat if scale_bits provided
        if self.scale_bits:
            float_max = tf.stop_gradient(max_value)
            return QFloat.quantize(inputs, float_max, self.value_bits, self.scale_bits)
        # Return a FixedPoint if no scale_bits provided
        frac_bits = tf.stop_gradient(self.frac_bits(max_value))
        return FixedPoint.quantize(inputs, self.value_bits, frac_bits)

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"scale_bits": self.scale_bits})
        return config


@tf.keras.utils.register_keras_serializable()
class OutputQuantizer(Calibrable, Quantizer):
    """A uniform quantizer that aligns its quantization range to a Power-of-two.
    Its max_value is updated during the calibration by the moving average algorithm,
    inspired by the BatchNormalizationLayer moving average algorithm.

    Args:
        bitwidth (int): the quantization bitwidth.
        signed (bool): whether the quantizer expects signed values or unsigned.
        momentum (float, optional): the momentum for the moving average. Defaults to 0.9.

    Note:
        To get more information about the moving average implementation, see the
        `BatchNormalizationLayer <https://bit.ly/3KcEaUh>`_ class:
    """

    def __init__(self, momentum=0.9, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.momentum = momentum
        # Add object that will store the shift values.
        self.shift = CalibrableVariable()

    def build(self, input_shape):
        """Build the layer.

        Args:
            input_shape (list): the shape of input tensor.
        """
        super().build(input_shape)

        # Declares the constant/vector that will store the maximum values of the input.
        with disable_partitioner(self):
            self.max_value = self.add_weight(
                name="max_value",
                shape=input_shape[-1] if self.axis is not None else (),
                dtype=tf.float32,
                initializer="ones",
                synchronization=tf.VariableSynchronization.ON_READ,
                trainable=False,
                aggregation=tf.VariableAggregation.MEAN,
                experimental_autocast=False,
            )

    @staticmethod
    def _assign_new_value(variable, value):
        """Given a variable, assign a new value to it. Function taken of
        `BatchNormalizationLayer <https://bit.ly/3v0gzll>`_ code.

        Args:
            variable (:obj:`tensorflow.Variable`): the variable to assign.
            value (:obj:`tensorflow.Tensor`): the new value to assign.

        Returns:
            :obj:`tensorflow.Tensor`: the new value of the variable.
        """
        with K.name_scope("AssignNewValue") as scope:
            # Expected match shape
            value_r = tf.reshape(value, tf.shape(variable))
            if tf.compat.v1.executing_eagerly_outside_functions():
                return variable.assign(value_r, name=scope)
            else:
                with tf.compat.v1.colocate_with(variable):
                    return tf.compat.v1.assign(variable, value_r, name=scope)

    @staticmethod
    def _assign_moving_average(variable, value, momentum, inputs_size):
        """Given a variable, assign a new value to it, using a moving average.
        Function taken of `BatchNormalizationLayer <https://bit.ly/3JUcLGd>`_ code.

        Args:
            variable (:obj:`tensorflow.Variable`): the variable to assign.
            value (:obj:`tensorflow.Tensor`): the new value to assign.
            momentum (float): the momentum for the moving average.
            inputs_size (int): the size of the inputs.

        Returns:
            :obj:`tensorflow.Tensor`: the new value of the variable.
        """

        def calculate_update_delta():
            decay = tf.convert_to_tensor(1.0 - momentum, name="decay")
            if decay.dtype != variable.dtype.base_dtype:
                decay = tf.cast(decay, variable.dtype.base_dtype)
            # Expected match shape
            value_r = tf.reshape(value, tf.shape(variable))
            update_delta = (variable - tf.cast(value_r, variable.dtype)) * decay
            if inputs_size is not None:
                update_delta = tf.where(
                    inputs_size > 0, update_delta, K.zeros_like(update_delta))
            return update_delta

        with K.name_scope("AssignMovingAvg") as scope:
            if tf.compat.v1.executing_eagerly_outside_functions():
                return variable.assign_sub(calculate_update_delta(), name=scope)
            else:
                with tf.compat.v1.colocate_with(variable):
                    return tf.compat.v1.assign_sub(variable, calculate_update_delta(), name=scope)

    def call(self, inputs, training=None):
        """Update the max_value from moving average, and quantize the inputs in three steps:

            1. Check that the inputs Tensor is a FixedPoint,
            2. Update the max_value from moving average, only if calibration is enabled and
            3. Quantize the inputs.

        Args:
            inputs (:obj:`tensorflow.Tensor` or :obj:`QTensor`): the inputs tensor.
            training (bool, optional): the training mode. Defaults to None.

        Returns:
            :obj:`tensorflow.Tensor`: the quantized tensor.
        """
        if not isinstance(inputs, FixedPoint):
            raise TypeError("The OutputQuantizer accepts only FixedPoint inputs."
                            f"Received {type(inputs)} inputs.")

        if inputs.value_bits <= self.value_bits:
            msg = f"Quantizing a {inputs.value_bits}-bit QTensor to "\
                f"{self.value_bits}-bit is pointless."
            if inputs.value_bits < self.value_bits:
                msg += " Use a promotion instead."
            raise ValueError(msg)

        if self.calibration:
            # Retrieve information from the inputs and update the weights
            input_batch_size = tf.shape(inputs)[0]
            if tf.reduce_all(tf.math.equal(self.max_value, tf.constant(1.))):
                momentum = tf.constant(-1.)
            else:
                momentum = tf.convert_to_tensor(self.momentum)
            # Compute the new value for all weights
            max_value = tf.math.reduce_max(tf.math.abs(inputs.to_float()), self.axis)
            # If max_values never updated set their newly computed values otherwise
            # update with moving average algorithm
            if momentum == -1:
                OutputQuantizer._assign_new_value(self.max_value, max_value)
            else:
                OutputQuantizer._assign_moving_average(
                    self.max_value, max_value, momentum, input_batch_size)

        # Quantize/downscale the input
        frac_bits = tf.stop_gradient(self.frac_bits(self.max_value))
        inputs, shift_value = inputs.downscale(frac_bits, self.value_bits)
        # update shift values
        self.shift(shift_value)
        return inputs

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"momentum": self.momentum})
        return config


@tf.keras.utils.register_keras_serializable()
class Dequantizer(Layer):
    """ Layer that allows to dequantize its inputs.
    """

    def call(self, inputs, training=None):
        """Convert QTensor inputs to float.

        Args:
            inputs (:obj:`tensorflow.Tensor` or :obj:`QTensor`): the inputs tensor(s).
            training (bool, optional): the training mode. Defaults to None.

        Returns:
            :obj:`tensorflow.Tensor`: the dequantized tensor(s).
        """
        def dequantize(x):
            if isinstance(x, QTensor):
                return x.to_float()
            return x

        if isinstance(inputs, (list, tuple)):
            return [dequantize(x) for x in inputs]

        return dequantize(inputs)



from .qtensor import *
from .qfloat import *
from .fixed_point import *
from .tf_dispatch import *
from .reciprocal import *
from .lookup import StaticHashTable, RegisterCustomHashTable

MAX_BUFFER_BITWIDTH = 24



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import keras
import tensorflow as tf

from ..tensors import ceil_log2, FixedPoint, MAX_BUFFER_BITWIDTH
from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, OutputQuantizer
from ..debugging import assert_less_equal


__all__ = ["Reciprocal", "QuantizedReciprocal"]


@tf.keras.utils.register_keras_serializable()
class Reciprocal(keras.layers.Layer):
    """Layer that computes the reciprocal of the input.
    """

    def call(self, inputs, training=False):
        return tf.math.reciprocal(inputs)


@tf.keras.utils.register_keras_serializable()
class QuantizedReciprocal(Calibrable, Reciprocal):
    """Piece-wise approximation of y = 1/x.

    The approximation works on values in the range [1, 2).
    To go into that range, we shift the input to put the leftmost bit (MSB) to
    1 and change the frac_bits so that there is only one int bit.
    Once that is done the approximation is as follows:

    y' = 1.59375 - 0.625 * x if x < 1.5
    y' = 1.125 - 0.3125 * x  if x >= 1.5

    Note that this can be performed in hardware this way:

    y' = 51 * 2^-5 - x * (2^-1 + 2^-3) if x < 1.5
    y' = 36 * 2^-5 - x * (2^-2 + 2^-4) if x >= 1.5

    Implementation inspired by:

    Cardarilli, G.C., Di Nunzio, L., Fazzolari, R. et al.
    A pseudo-softmax function for hardware-based high speed image classification.
    Sci Rep 11, 15307 (2021). https://doi.org/10.1038/s41598-021-94691-7

    Args:
        quant_config (dict, optional): the quantization configuration. Defaults to {}.

    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.buffer_bitwidth = quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        self.out_quantizer = deserialize_quantizer(
            quant_config, "output_quantizer", OutputQuantizer, False)

        # All constants are in the range |x| < 2. We only need 1 bit for the int part.
        # However, two approaches are considered for the fractional part:
        # 1. 'Multiplying' constants will have the smallest possible frac_bits (4)
        mul_frac_bits = mul_value_bits = 4
        self._K_0_3125 = FixedPoint.quantize(-0.3125, mul_value_bits, mul_frac_bits)
        self._K_0_625 = FixedPoint.quantize(-0.625, mul_value_bits, mul_frac_bits)
        # 2. 'Adding' constant will be aligned with the result of x' = c*x (c as constant)
        #    Note: it is possible to calculate a-priori the alignment, since algorithm ensures
        #    that x'.frac_bits = x.frac_bits + c.frac_bits = (self.buffer_bitwidth - 1 - 4) + 4
        #                      = self.buffer_bitwidth - 1
        value_bits = self.buffer_bitwidth
        frac_bits = value_bits - 1
        self._K_1_125 = FixedPoint.quantize(1.125, value_bits, frac_bits)
        self._K_1_59375 = FixedPoint.quantize(1.59375, value_bits, frac_bits)

        # The limit is to choose the PWL function is 1.5 (= 3 * 2^-1), that only containts
        # 1 bit in int part. But, this comparison happens before to any mathematical operation,
        # so we need that self._limit_1_5.frac_bits = x.frac_bits to avoid implicit aligmnments
        self._limit_1_5 = FixedPoint.quantize(1.5, value_bits, frac_bits - mul_frac_bits)

        # We need to store the input frac bits to reproduce the reciprocal on hardware.
        self.input_frac_bits = CalibrableVariable()

    def _reciprocal_x_ge_1_5_values(self, x):
        """Implements reciprocal approximation when x in [1.5, 2.0), following the equation:

        1/x = 1.125 - 0.3125 * x

        Args:
            x (:obj:`FixedPoint`): the value to compute its reciprocal.

        Returns:
            :obj:`FixedPoint`: the reciprocal result.
        """
        x = x * self._K_0_3125
        x_reciprocal = x + self._K_1_125
        return x_reciprocal.values

    def _reciprocal_x_lt_1_5_values(self, x):
        """Implements reciprocal approximation when x in [1.0, 1.5), following the equation:

        1/x = 1.59375 - 0.625 * x

        Args:
            x (:obj:`FixedPoint`): the value to compute its reciprocal.

        Returns:
            :obj:`FixedPoint`: the reciprocal result.
        """
        x = x * self._K_0_625
        x_reciprocal = x + self._K_1_59375
        return x_reciprocal.values

    def call(self, inputs, training=False):
        if not isinstance(inputs, FixedPoint):
            raise TypeError("QuantizedReciprocal only accepts FixedPoint inputs. "
                            f"Receives '{type(inputs)}' as inputs.")

        # In multiplication constant, value_bits = frac_bits and they are fixed to 4.
        # We take value_bits to avoid going through tensorflow.
        mul_frac_bits = self._K_0_3125.value_bits

        # Promote and get properties of FixedPoint
        inputs = inputs.promote(self.buffer_bitwidth)
        x_values = inputs.values
        x_frac_bits = inputs.frac_bits
        x_value_bits = inputs.value_bits

        # To avoid saturation in multiplication, we will use a "fake buffer",
        # limited to 'mul_frac_bits' bits less that the original.
        x_aligned_bits = x_value_bits - mul_frac_bits

        # Store input frac_bits
        self.input_frac_bits(x_frac_bits)

        # Evaluate element-wise the number of bits used in x binary representation.
        # This operation can be done in hardware by counting leading zeros.
        used_bits = ceil_log2(x_values)
        # This operation only have sense if x_mul_value_bits >= used_bits
        assert_less_equal(used_bits, x_aligned_bits, f"required bits ({used_bits}) exceed "
                          f"the buffer ({x_aligned_bits}).")

        # The PWL function works in the range [1, 2).
        #
        # To project x into that range, we first align it element-wise to make sure
        # all items binary representation leftmost-bit are at one.
        #
        # x' = x << (x_aligned_bits - used_bits)
        #
        # where x_aligned_bits = x_value_bits - mul_frac_bits, which will project the
        # values of x to a maximum of x_aligned_bits. That means:
        #
        # x'.values = x.values * 2 ^ (x_aligned_bits - used_bits)
        # x'.frac_bits = x_frac_bits + (x_aligned_bits - used_bits)
        #
        # Now, if we define y as the number with the same values as x' but (x_aligned_bits - 1)
        # fractional bits, it is guaranteed it is in the interval [1, 2).
        #
        # y.values = x'.values = x.values * 2 ^ (x_aligned_bits - used_bits)
        # y.frac_bits = x_aligned_bits - 1
        #
        y_values = inputs.shift(x_aligned_bits - used_bits).values
        y_frac_bits = x_aligned_bits - 1
        y = FixedPoint(y_values, y_frac_bits, x_value_bits)

        # Estimate output values using one of the PWL functions
        reciprocal_y_values = tf.where(y >= self._limit_1_5,
                                       self._reciprocal_x_ge_1_5_values(y),
                                       self._reciprocal_x_lt_1_5_values(y))
        # After the approximation, we know that the fractional bits has increased
        # 'mul_frac_bits' times, therefore:
        # y_frac_bits' = y_frac_bits + mul_frac_bits
        #              = (x_aligned_bits - 1) + mul_frac_bits
        #              = (x_value_bits - mul_frac_bits) - 1 + mul_frac_bits
        #              = x_value_bits - 1
        # This is sure given the fact that the range of reciprocal_y is within (0,1)
        y_frac_bits = x_value_bits - 1
        reciprocal_y = FixedPoint(reciprocal_y_values, y_frac_bits, x_value_bits)

        # Since we defined previously
        #
        # y.values = x.values * 2 ^ (x_aligned_bits - used_bits)
        # y.frac_bits = x_aligned_bits - 1
        #
        # We can express y as:
        #
        # y = y.values * 2 ^ (1 - x_aligned_bits)
        # y = x.values * 2 ^ (x_aligned_bits - used_bits) * 2 ^ (1 - x_aligned_bits)
        # y = x.values * 2 ^ (1 - used_bits)
        #
        # or x.values = y * 2 ^ (used_bits - 1)
        #
        # We can now express x as:
        # x = x.values * 2 (-x.frac_bits)
        # x = y * 2 ^(used_bits - 1 - x.frac_bits)
        #
        # and:
        #
        # 1/x = 1/y * 2^(x.frac_bits - used_bits + 1)
        #
        outputs = reciprocal_y.shift(x_frac_bits - used_bits + 1)

        # Finally, passes result through output_quantizer
        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import keras
import tensorflow as tf


__all__ = ["Calibrable", "CalibrableVariable"]


class Calibrable():
    """A class that exhibits a 'calibration' property.

    All objects inheriting from this class share the same 'calibration' property.

    Setting the property on one instance will automatically set the property
    on all members of the class that are also Calibrable objects.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._calibration = False
        self.calibrables = []

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if isinstance(value, Calibrable):
            self.calibrables.append(name)

    @property
    def calibration(self):
        """Flag to specify if the object is in calibration mode or not.

        Returns:
            bool: True if the object is in calibration mode, False otherwise.
        """
        return self._calibration

    @calibration.setter
    def calibration(self, value):
        """Set the calibration mode of the object, and all its members.

        Args:
            value (bool): True if the object will be in calibration mode, False otherwise.
        """
        self._calibration = value
        for name in self.calibrables:
            calibrable = getattr(self, name)
            calibrable.calibration = value


class CalibrableVariable(Calibrable, keras.layers.Layer):
    """Wrapper class to store and retrieve a calibrable variable, e.g.: shift
    information.
    """

    def build(self, input_shape):
        super().build(input_shape)
        self._var = self.add_weight(
            shape=input_shape,
            dtype=tf.float32,
            initializer="zeros",
            synchronization=tf.VariableSynchronization.ON_READ,
            trainable=False,
            aggregation=tf.VariableAggregation.MEAN,
            experimental_autocast=False,
        )

    @property
    def value(self):
        """Get to the variable value.

        Returns:
            :obj:`tf.Tensor`: value of the stored variable.
        """
        return self._var.value()

    def call(self, inputs):
        """Updates the value of the variable if calibration is True.

        Args:
            inputs (:obj:`tf.Tensor`): new values.

        Returns:
            :obj:`tf.Tensor`: value of the variable when calibration is True, None otherwise.
        """
        if not self.calibration:
            return None
        self._var.assign(inputs)
        return inputs



from .activations import *
from .calibrable import *
from .dense import *
from .dropout import *
from .layers import *
from .normalization import *
from .quantizers import *
from .shiftmax import *
from .vit import *
from .depthwise_convolution import *
from .convolution import *
from .separable_convolution import *
from .reshaping import *
from .attention import *
from .pooling import *
from .reciprocal import *



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
import keras

from ...debugging import assert_rank_at_least
from ..qtensor import QTensor
from ..fixed_point import FixedPoint
from ..qfloat import QFloat


@tf.experimental.dispatch_for_api(tf.linalg.matmul)
def fp_matmul(a: FixedPoint,
              b: QTensor,
              transpose_a=False,
              transpose_b=False,
              adjoint_a=False,
              adjoint_b=False,
              a_is_sparse=False,
              b_is_sparse=False,
              output_type=None,
              name=None):
    """Multiplies matrix `a` by matrix `b`, producing `a` * `b`.

    A `tf.Tensor` of the same type as `a` and `b` where each inner-most matrix
    is the product of the corresponding matrices in `a` and `b`, e.g. if all
    transpose or adjoint attributes are `False`:
    `output[..., i, j] = sum_k (a[..., i, k] * b[..., k, j])`,
    for all indices `i`, `j`.

    Note: This is matrix product, not element-wise product.


    Args:
        a (:obj:`FixedPoint`): a FixedPoint of rank > 1.
        b (:obj:`FixedPoint`): a FixedPoint with same rank as `a`.
        transpose_a (bool, optional): if `True`, `a` is transposed before multiplication.
            Defaults to False.
        transpose_b (bool, optional): if `True`, `b` is transposed before multiplication.
            Defaults to False.
        adjoint_a (bool, optional): if `True`, `a` is conjugated and transposed before
            multiplication. Defaults to False.
        adjoint_b (bool, optional): if `True`, `b` is conjugated and transposed before
            multiplication. Defaults to False.
        a_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        b_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        output_type (NoneType, optional): must be None, argument kept for compatibility
            with original tf.matmul. Defaults to None.
        name (str, optional): the name for the operation. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the multiplied FixedPoint.
    """
    if a_is_sparse:
        raise TypeError(
            f"Unsupported argument: a_is_sparse, provided {a_is_sparse}")
    if b_is_sparse:
        raise TypeError(
            f"Unsupported argument: b_is_sparse, provided {b_is_sparse}")
    if output_type is not None:
        raise TypeError(
            f"Unsupported argument: output_type, provided {output_type}")

    # We don't support matmul between vectors
    assert_rank_at_least(a.values, 2)
    assert_rank_at_least(b.values, 2)

    # Since the last dimension is collapsed by the matmul, (a,b) must be quantized per-tensor
    if not transpose_a:
        a.assert_per_tensor()
    if transpose_b:
        b.assert_per_tensor()

    if isinstance(b, FixedPoint):
        b_frac_bits = b.frac_bits
    else:
        b_frac_bits = b.scales.frac_bits

    # Do matmul on values
    m_values = tf.matmul(a.values, b.values, transpose_a, transpose_b, adjoint_a, adjoint_b, name)
    if isinstance(b, QFloat):
        # Multiply by the scales
        m_values *= b.scales.values

    # Return a new FixedPoint
    return FixedPoint(m_values, a.frac_bits + b_frac_bits, a.value_bits)


@tf.experimental.dispatch_for_api(keras.backend.depthwise_conv2d)
def fp_depthwise_conv2d(x: FixedPoint,
                        depthwise_kernel: QFloat,
                        strides=(1, 1),
                        padding='valid',
                        data_format=None,
                        dilation_rate=(1, 1)):
    """ 2D convolution with separable filters.

    Args:
        x (obj:`FixedPoint`): input tensor.
        depthwise_kernel (:obj:`QFloat`): convolution kernel for the depthwise convolution.
        strides (tuple, optional): strides tuple (length 2). Defaults to (1, 1).
        padding (str, optional): `"same"` or `"valid"`. Defaults to 'valid'.
        data_format (str, optional): `"channels_last"` or `"channels_first"`. Defaults to None.
        dilation_rate (tuple, optional): tuple of integers, dilation rates for the separable
            convolution. Defaults to (1, 1).

    Returns:
        :obj:`FixedPoint`: output tensor.

    """
    # Unlike other convolutions, the depthwise does not require its inputs to
    # be quantized per-tensor as the input channels are processed independently

    # Do convolution on values
    conv_values = keras.backend.depthwise_conv2d(x.values, depthwise_kernel.values, strides,
                                                 padding, data_format, dilation_rate)
    # Multiply by the scales
    conv_values *= depthwise_kernel.scales.values

    # Return a new FixedPoint
    filters_frac_bits = depthwise_kernel.scales.frac_bits
    return FixedPoint(conv_values, x.frac_bits + filters_frac_bits, x.value_bits)


@tf.experimental.dispatch_for_api(tf.nn.convolution)
def fp_convolution(
        input: FixedPoint,
        filters: QFloat,
        strides=None, padding="VALID", data_format=None, dilations=None, name=None):
    """Perform the convolution operation between input and filter tensors.

    Args:
        input (:obj:`FixedPoint`): The input FixedPoint.
        filters (:obj:`QFloat`): The filters Qtensor.
        strides (list, optional): Sequence of N ints >= 1.  Specifies the output stride.
        padding (str, optional): A string, either `"VALID"` or `"SAME"`. The padding algorithm.
            Defaults to "VALID"
        data_format (str, optional): Specifies whether the channel dimension of
            the `input` and output is the last dimension (default, or if `data_format`
            does not start with "NC"), or the second dimension (if `data_format`
            starts with "NC").  For N=1, the valid values are "NWC" (default) and
            "NCW".  For N=2, the valid values are "NHWC" (default) and "NCHW".
            For N=3, the valid values are "NDHWC" (default) and "NCDHW".
        dilations (list, optional): Alias for dilation_rate. Sequence of N ints >= 1.
            Specifies the filter upsampling/input downsampling rate. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: a FixedPoint containing the output values.
    """
    # Input must be quantized per-tensor because the products are eventually summed
    input.assert_per_tensor()

    # Do convolution on values
    conv_values = tf.nn.convolution(input.values, filters.values, strides, padding, data_format,
                                    dilations, name)
    # Multiply by the scales
    conv_values *= filters.scales.values

    # Return a new FixedPoint
    filters_frac_bits = filters.scales.frac_bits
    return FixedPoint(conv_values, input.frac_bits + filters_frac_bits, input.value_bits)
